"""Module for conveniently managing a Hyperopt cluster. https://github.com/hyperopt/hyperopt
"""

import time
from typing import List, Union, Optional
import os

from lazycluster import RuntimeTask, Runtime, RuntimeGroup
from lazycluster.cluster import MasterWorkerCluster, MasterLauncher, WorkerLauncher
from lazycluster import _utils
from lazycluster.cluster.exceptions import MasterStartError
from lazycluster.exceptions import PortInUseError
from lazycluster.utils import Environment


class LocalMongoLauncher(MasterLauncher):
    """Concrete implementation of the `MasterLauncher` interface. See its documentation to get a list of the inherited
    methods and attributes.

    This class implements the logic for starting a MongoDB instance on localhost. Hence, we simply treat the MongoDB
    instance as master node.
    """

    def __init__(self, runtime_group: RuntimeGroup):
        """Initialization method.

        Args:
            runtime_group: The group where the workers will be started.
        """
        super().__init__(runtime_group)
        self.dbpath = None

    def start(self, ports: Union[List[int], int], timeout: int = 0, debug: bool = False) -> List[int]:
        """Launch a master instance.

        Note:
            If you create a custom subclass of MasterLauncher which will not start the master instance on localhost
            then you should pass the debug flag on to `execute_task()` of the `RuntimeGroup` or `Runtime` so that you
            can benefit from the debug feature of `RuntimeTask.execute()`.

        Args:
            ports: Port where the DB should be started. If a list is given then the first port that is free in the
                   `RuntimeGroup` will be used. The actual chosen port can be requested via the property `port`.
            timeout: Timeout (s) after which an MasterStartError is raised if DB instance not started yet. Defaults to
                     3 seconds.
            debug: If `True`, stdout/stderr from the runtime will be printed to stdout of localhost. If, `False` then
                   the stdout/stderr will be added to python logger with level debug after each `RuntimeTask` step. Defaults to
                   `False`.

        Returns:
            List[int]: In case a port list was given the updated port list will be returned. Otherwise an empty list.

        Raises:
            PortInUseError: If a single port is given and it is not free in the `RuntimeGroup`.
            NoPortsLeftError: If a port list was given and none of the ports is actually free in the `RuntimeGroup`.
            MasterStartError: If master was not started after the specified `timeout`.
        """

        if debug:
            self.log.debug('The debug flag has no effect in LocalMongoLauncher.')

        if not isinstance(ports, list):
            if _utils.localhost_has_free_port(ports) and \
               self._group.has_free_port(ports, exclude_hosts=Runtime.LOCALHOST):
                self._port = master_port = ports
            else:
                raise PortInUseError(ports, self._group)

        else:
            self._port = master_port = self._group.get_free_port(ports)  # Raises NoPortsLeftError
            ports = _utils.get_remaining_ports(ports, master_port)

        self.log.debug(f'Starting MongoDB on localhost on port {str(master_port)} with dbpath `{self.dbpath}` and '
                       f'logfile `{self.dbpath}/{HyperoptCluster.MONGO_LOG_FILENAME}`.')

        # Start the mongod deamon process
        return_code = os.system(self.get_mongod_start_cmd())

        if return_code != 0:
            cause = f'Please verify that (1) MongoDB is installed, (2) the dbpath `{self.dbpath}` exists with the ' \
                    f'rights required by mongod and (3) that no other MongoDB instance is using and consequently ' \
                    f'locking the respective files (=> Init HyperoptCluster with another dbpath ' \
                    f'or manually stop the mongod process). See hyperopt docs in README for further details.'
            raise MasterStartError('localhost', master_port, cause)

        time.sleep(timeout)  # Needed for being able to check the port

        if not _utils.localhost_has_free_port(master_port):
            self.log.info('MongoDB started on localhost on port ' + str(self._port))
        else:
            self.log.debug('MongoDB could NOT be started successfully on port ' + str(self._port))
            cause = f'The master port {master_port} is still free when checking after the timeout of {timeout} seconds.'
            raise MasterStartError('localhost', master_port, cause)

        # Sets up ssh tunnel for scheduler such that all communication is routed over the
        # local machine and all entities can talk to each the scheduler on localhost.
        self.log.info(f'Expose the MongoDB port in the RuntimeGroup.')
        self._group.expose_port_to_runtimes(self._port)

        return ports

    def get_mongod_start_cmd(self) -> str:
        """Get the shell command for starting mongod as a deamon process.

            Returns:
                str: The shell command.
        """
        return f'mongod --fork --logpath={self.dbpath}/{HyperoptCluster.MONGO_LOG_FILENAME} --dbpath={self.dbpath} ' \
               f'--port={self._port}'

    def get_mongod_stop_cmd(self) -> str:
        """Get the shell command for stopping the currently running mongod process.

            Returns:
                str: The shell command.
        """
        return f'mongod --shutdown --dbpath={self.dbpath}'

    def cleanup(self):
        """Release all resources.
        """
        self.log.info('Stop the MongoDB ...')
        self.log.debug('Cleaning up the LocalMasterLauncher ...')
        return_code = os.system(self.get_mongod_stop_cmd())
        if return_code == 0:
            self.log.info('MongoDB successfully stopped.')
        else:
            self.log.warning('MongoDB daemon could NOT be stopped.')
        super().cleanup()


class RoundRobinLauncher(WorkerLauncher):
    """Concrete WorkerLauncher implementation for launching hyperopt workers in a round robin manner. See its
    documentation to get a list of the inherited methods and attributes.
    """

    def __init__(self, runtime_group: RuntimeGroup, dbname: str, poll_interval: float):
        """Initialization method.

        Args:
            runtime_group: The group where the workers will be started.
            dbname: The name of the mongodb instance.
            poll_interval: The poll interval of the hyperopt worker.

        Raises.
            ValueError: In case dbname is empty.
        """
        super().__init__(runtime_group)
        self._ports = None

        if not dbname:
            raise ValueError('dbname must not be empty')
        self._dbname = dbname

        if not poll_interval:
            raise ValueError('poll_interval must not be empty')
        self._poll_interval = poll_interval

        self.log.debug('RoundRobinLauncher initialized.')

    def start(self, worker_count: int, master_port: int, ports: List[int] = None, debug: bool = True) -> List[int]:
        """Launches the worker instances in the `RuntimeGroup`.

        Args:
            worker_count: The number of worker instances to be started in the group.
            master_port:  The port of the master instance.
            ports: Without use here. Only here because we need to adhere to the interface defined by the
                   WorkerLauncher class.
            debug: If `True`, stdout/stderr from the runtime will be printed to stdout of localhost. If, `False` then
                   the stdout/stderr will be added to python logger with level debug after each `RuntimeTask` step. Defaults to
                   `False`.

        Returns:
            List[int]: The updated port list after starting the workers, i.e. the used ones were removed.
        """
        self._ports = ports
        hosts = self._group.hosts
        runtimes = self._group.runtimes

        # Launch each desired worker one by one
        for worker_index in range(worker_count):
            # Determine the runtime where the next worker will be started in
            runtime_index = (self._group.runtime_count + worker_index) % self._group.runtime_count
            # Get the actual host corresponding to the index
            host = hosts[runtime_index]
            # working_dir = runtimes[runtime_index].working_dir
            assert host == runtimes[runtime_index].host

            self.log.debug(f'Launch Hyperopt worker with index {worker_index} on Runtime {host}')

            self._launch_single_worker(host, worker_index, master_port, debug)

        return self._ports

    def _launch_single_worker(self, host: str, worker_index: int, master_port: int, debug):
        """Launch a single worker instance in a `Runtime` in the `RuntimeGroup`.
        """
        # 2. Start the worker on this port
        task = RuntimeTask('launch-hyperopt-worker-' + str(worker_index))
        task.run_command(self._get_launch_command(master_port, self._dbname, self._poll_interval))
        self._group.execute_task(task, host, omit_on_join=True, debug=debug)

    @classmethod
    def _get_launch_command(cls, master_port: int, dbname: str, poll_interval: float = 0.1) -> str:
        """Get the shell command for starting a worker instance.

        Returns:
             str: The launch command.
        """
        return f'hyperopt-mongo-worker --mongo=localhost:{str(master_port)}/{dbname} ' \
               f'--poll-interval={str(poll_interval)}'

    def cleanup(self):
        """Release all resources.
        """
        self.log.info('Cleanup the RoundRobinLauncher ...')
        super().cleanup()


class HyperoptCluster(MasterWorkerCluster):
    """Convenient class for launching a Hyperopt cluster in a `RuntimeGroup`.

    HyperoptCluster inherits from MasterWorkerCluster. See its documentation to get a list of the inherited methods
    and attributes.

    The number of hyperopt workers defaults to the number of `Runtimes` in the used `RuntimeGroup`. This number can be
    adjusted so that more or less workers than available `Runtimes` can be used. Per default the desired number of
    workers is started in a round robin way as implemented in `RoundRobinLauncher`. Consequently, this leads to an
    equal distribution of hyperopt workers in the `RuntimeGroup`. You can provide a custom implementation inheriting
    from the `WorkerLauncher` class in order to execute a different strategy how workers should be started. The
    master instance (i.e. the mongoDB) will always be started on localhost as implemented in `LocalMasterLauncher`. This
    behavior can also be changed by providing a custom implementation inheriting from the `MasterLauncher`.
    """

    MONGO_LOG_FILENAME = 'hyperopt_mongo.log'
    DEFAULT_MASTER_PORT = 27017
    ENV_NAME_MONGO_URL = 'MONGO_CONNECTION_URL'

    def __init__(self, runtime_group: RuntimeGroup,
                 master_launcher: Optional[MasterLauncher] = None,
                 worker_launcher: Optional[WorkerLauncher] = None,
                 dbpath: Optional[str] = None,
                 dbname: str = 'hyperopt',
                 worker_poll_intervall: float = 0.1):
        """Initialization method.

        Args:
            runtime_group: The `RuntimeGroup` contains all `Runtimes` which can be used for starting the entities.
            master_launcher: Optionally, an instance implementing the `MasterLauncher` interface can be given, which
                             implements the strategy for launching the master instances in the cluster. If None, then
                             `LocalMasterLauncher` is used.
            worker_launcher: Optionally, an instance implementing the `WorkerLauncher` interface can be given, which
                             implements the strategy for launching the worker instances. If None, then
                             `RoundRobinLauncher` is used.
            dbpath: The directory where the db files will be kept. Defaults to a `mongodb` directory inside the
                    `utils.Environment.main_directory`.
            dbname: The name of the database to be used for experiments. See MongoTrials url scheme in hyperopt
                    documentation for more details. Defaults to ´hyperopt´.
            worker_poll_intervall: The poll interval of the hyperopt worker. Defaults to `0.1`.

        Raises:
            PermissionError: If the `dbpath` does not exsist and could not be created due to lack of permissions.
        """
        super().__init__(runtime_group)

        self._master_launcher = master_launcher if master_launcher else LocalMongoLauncher(runtime_group)

        if not dbpath:
            self._master_launcher.dbpath = os.path.join(Environment.main_directory, 'mongodb')
            try:
                os.makedirs(self._master_launcher.dbpath)  # Raises PermissionError
            except FileExistsError:
                pass  # All good because the dir already exists
        else:
            self._master_launcher.dbpath = dbpath
        self._dbname = dbname

        self._worker_launcher = worker_launcher if worker_launcher else RoundRobinLauncher(runtime_group, dbname,
                                                                                           worker_poll_intervall)

        self.log.debug('HyperoptCluster initialized.')

    @property
    def mongo_trial_url(self) -> str:
        """The MongoDB url indicating what mongod process and which database to use.

        Note:
            The format is the format required by the hyperopt MongoTrials object.

        Returns:
            str: URL string.
        """
        if not self.master_port:
            self.log.warning('HyperoptCluster.mongo_trial_url was requested although the master_port is not yet set.')

        return f'mongo://localhost:{self.master_port}/{self.dbname}/jobs'

    @property
    def mongo_url(self) -> str:
        """The MongoDB url indicating what mongod process and which database to use.

        Note:
            The format is `mongo://host:port/dbname`.

        Returns:
            str: URL string.
        """
        if not self.master_port:
            self.log.warning('HyperoptCluster.mongo_trial_url was requested although the master_port is not yet set.')

        return f'mongo://localhost:{self.master_port}/{self.dbname}'

    @property
    def dbname(self):
        """The name of the MongoDB database to be used for experiments.
        """
        return self._dbname

    def start_master(self, master_port: Optional[int] = None, timeout: int = 3, debug: bool = False):
        """Start the master instance.

        Note:
            How the master is actually started is determined by the the actual `MasterLauncher` implementation. Another
            implementation adhering to the `MasterLauncher` interface can be provided in the constructor of the cluster
            class.

        Args:
            master_port: Port of the master instance. Defaults to self.DEFAULT_MASTER_PORT, but another one is chosen if
                         the port is not free within the group. The actual chosen port can be requested via
                         self.master_port.
            timeout: Timeout (s) after which an MasterStartError is raised if master instance not started yet.
            debug: If `True`, stdout/stderr from the runtime will be printed to stdout of localhost. Has no effect for
                   if the master instance is started locally, what default MasterLauncher implementations usually do.

        Raises:
            PortInUseError: If a single port is given and it is not free in the `RuntimeGroup`.
            NoPortsLeftError: If there are no free ports left in the port list for instantiating the master.
            MasterStartError: If master was not started after the specified `timeout`.
        """
        super().start_master(master_port, timeout)
        self._group.add_env_variables({self.ENV_NAME_MONGO_URL: self.mongo_trial_url})

    def cleanup(self):
        """Release all resources.
        """
        self.log.info('Shutting down the HyperoptCluster...')
        super().cleanup()



import logging
import sys
import os

import click

from typing import Optional, List
import storm.__main__ as storm

from click_spinner import spinner
from lazycluster import RuntimeManager, NoRuntimesDetectedError

log = logging.getLogger(__name__)


@click.group()
@click.version_option()
@click.option('--debug/--no-debug', '-d', default=False)
def cli(debug: bool):
    # log to sys out
    log_level = logging.DEBUG if debug else logging.CRITICAL
    logging.basicConfig(stream=sys.stdout, format='%(asctime)s : %(levelname)s : %(message)s', level=logging.CRITICAL)
    logging.getLogger('paramiko').setLevel(logging.CRITICAL)
    logging.getLogger('lazycluster').setLevel(log_level)


@cli.command('add-runtime')
@click.argument("name")
@click.argument("connection_uri")
@click.option("--id_file", "-id", required=False, type=click.STRING,
              help="The private key file that should be used for authentication")
@click.option("--options", "-o", required=False, default=[], help="Custom ssh options")
@click.option("--config", "-c", required=False, type=click.STRING, help="The ssh config file")
def add_runtime(name: str, connection_uri: str, id_file: Optional[str] = None, options: List = [], config: Optional[str] = None):
    storm.add(name, connection_uri, id_file, options, config)


@cli.command('delete-runtime')
@click.argument("name")
@click.option("--config", "-c", required=False, type=click.STRING, help="The ssh config file")
def delete_runtimes(name: str, config: Optional[str] = None):
    # Delete the related ssh config
    storm.delete(name, config)

    # Delete configured remote kernel if present
    try:
        import remote_ikernel

        kernel_name = 'rik_ssh_' + name.replace('-', '_') + '_py36'

        os.system('remote_ikernel manage --delete ' + kernel_name)

        # The Identity file won't be removed. Maybe it is useful to provide an additional flag to enforce the removal
        # can be retrieved from storm via single_ssh_entry['options']['identityfile']
    except Exception:
        pass

    print('Runtime successfully deleted.')


@cli.command('list-runtimes')
@click.option("--long", "-l", is_flag=True, help="Print detailed information about the Runtimes")
def list_runtime(long: bool):
    with spinner():

        try:
            runtime_mgr = RuntimeManager()
        except NoRuntimesDetectedError:
            print('\nNo runtimes detected!\n')
            return

        if long:
            # Ensure that all runtime info are read asynchronously before printing.
            # Otherwise each runtime will read its information synchronously.
            runtime_mgr._group.fill_runtime_info_buffers_async()

    if long:
        runtime_mgr.print_runtime_info()
    else:
        # Only printing the hosts is much faster than reading the detailed host information
        runtime_mgr.print_hosts()


@cli.command('start-dask')
def start_dask_cluster():
    from lazycluster.cluster.dask_cluster import DaskCluster
    cluster = DaskCluster(RuntimeManager().create_group())
    cluster.start()
    while True:
        pass


if __name__ == '__main__':
    cli()


from __future__ import absolute_import

from lazycluster.cluster.runtime_cluster import MasterLauncher, WorkerLauncher, RuntimeCluster, MasterWorkerCluster

from lazycluster.cluster.dask_cluster import DaskCluster
from lazycluster.cluster.hyperopt_cluster import HyperoptCluster
#from lazycluster.cluster.ray_cluster import RayCluster


"""Contains setting parameters for the library. """

PIP_PROJECT_NAME = 'lazycluster'
GITHUB_URL = 'https://github.com/ml-tooling/lazycluster.git'
BRANCH = 'develop'  # Leave empty to enforce lazycluster installation from PyPi - see _utils.get_pip_install_cmd()


"""Runtimes module.

This module comprises classes for executing so called `RuntimeTasks` in `Runtimes` by leveraging the power of ssh.
The `RuntimeTask` class is a container for defining a sequence of elegantly task steps. This `RuntimeTask` can then be
executed either standalone of by passing it over to a `Runtime` instance. Passwordless ssh should be configured for all
hosts that should act as a `Runtime` to be able to conveniently manage those entities.

"""
import tempfile
import time

from fabric import Connection
from multiprocessing import Process, Manager
from typing import Optional, List, Dict, Generator, Union
import shutil
import cloudpickle as pickle
import atexit
from lazycluster import InvalidRuntimeError, NoPortsLeftError, PortInUseError, TaskExecutionError, PathCreationError
import lazycluster._utils as _utils
from lazycluster.utils import ExecutionFileLogUtil

import warnings
import logging
import os


class RuntimeTask(object):
    """This class provides the functionality for executing a sequence of elementary operations over ssh. The [fabric](http://docs.fabfile.org/en/2.5/api/connection.html)
    library is used for handling ssh connections. A `RuntimeTask` can be composed from four different operations which
    we call steps, namely adding a step for running a shell command via `run_command()`, sending a file to a host via
    `send_file()`, retrieving a file from a host via `get_file()` or adding a step for executing a python function on a
    host via `run_function()`. The current restriction for running functions is that these functions need to be
    serializable via cloudpickle. To actually execute a `RuntimeTask`, i.e. the sequence of task steps, either a call
    to `execute()` is necessary or a handover to the `execute_task()` method of the `Runtime` class is necessary.
    Usually, a `RuntimeTask` or `RuntimeGroup` will be executed in a `Runtime` or in a `RuntimeGroup`. See its documentation for further
    details.

    Examples:
        ```python
        # 1. Define a function that should be executed remotely via a RuntimeTask
        def print():
            print('Hello World!')

        # 2. Create & compose the RuntimeTask by using the elementary operations
        my_task = RuntimeTask('my-task').run_command('echo Hello World!').run_function(print)

        # 3. Execute the RuntimeTask standalone w/o Runtime by handing over a fabric ssh connection
        from fabric import Connection
        task = my_task.execute(Connection('host'))

        # 4. Check the logs of the RuntimeTask execution
        task.print_log()
        log = task.execution_log
        ```
    """

    def __init__(self, name: Optional[str] = None):
        """Initialization method.
        
        Args:
            name: The name of the task. Defaults to None and consequently a unique identifier is generated via Python's
                  id() function.
        """

        # Create the Logger
        self.log = logging.getLogger(__name__)
        self._execution_log_file_path = None

        self.name = name
        if not self.name:
            self.name = str(id(self))

        self._task_steps = []
        self._execution_log = []
        self._function_return_pkl_paths = []  # file paths to files with function's pickled return data
        self._process = None

        self.omit_on_join = False

        self._env_variables = {}  # will be passed on to the fabric connection

        # Will be created if run_function is executed      
        self._temp_dir = None

        # This indicates how many deep copies were created based on this task.
        # E.g. if this task was deep copied tiwce, then the index would be 2.
        self._copy_index = 0

        # Cleanup will be done atexit since usage of destructor may lead to exceptions
        atexit.register(self.cleanup)

        self.log.debug(f'RuntimeTask {self.name} initialized.')

    def __repr__(self):
        return "%s(%r)" % (self.__class__, self.__dict__)

    def __str__(self):
        return type(self).__name__ + ': ' + self.name

    def __deepcopy__(self, memo: Optional[dict] = None):
        """This functions implements the deep copy logic so that all relevant data is copied or recreated with similar
        values.

        Note:
            A custom implementation is necessary here since especially in run_function we automatically generate  pickle
            file paths. When broadcasting a task we need to copy the task since it holds state such as logs and paths.
            In order to circumvent the overwriting of such files, we need to ensure that new paths must be created.
        """
        # Increment the task index and append the index to the end of the task name
        # -> simplifies debugging
        self._copy_index += 1
        new_name = f'{self.name}-{self._copy_index}'

        copied_task = RuntimeTask(new_name)

        # Copy the task steps and re-execute run_function on the copied task,
        # since there will be individual file paths created among others.
        for step in self._task_steps:
            if step.type != self._TaskStep.TYPE_RUN_FUNCTION:
                copied_task._task_steps.append(step)
            else:
                copied_task.run_function(step.function, **step.func_kwargs)

        copied_task._env_variables = self._env_variables
        copied_task.omit_on_join = self.omit_on_join

        self.log.debug(f'Deep copy [{self._copy_index}] of RuntimeTask {self.name} created by using its custom '
                       f'__deepcopy__ implementation.')

        return copied_task

    def cleanup(self):
        """Remove temporary used resources, like temporary directories if created.
        """
        self.log.debug(f'Start cleanup of task {self.name}.')
        if self._temp_dir:
            shutil.rmtree(self._temp_dir)
            self.log.debug(f'Temporary directory {self._temp_dir} of RuntimeTask {self.name} on localhost removed.')
            self._temp_dir = None

    @property
    def execution_log(self) -> List[str]:
        """The execution log as list. The list is empty as long as a task was not yet executed. Each log entry
        corresponds to a single task step and the log index starts at `0`. If th execution of an individual step does not
        produce and outut the list entry will be empty.
        """
        return self._execution_log

    @property
    def execution_log_file_path(self) -> str:
        """The execution log file path. This property is read-only and
        will be updated each time the `RuntimeTask` gets executed.
        """
        return self._execution_log_file_path
    @property
    def function_returns(self) -> Generator[object, None, None]:
        """The return data produced by functions which were executed as a consequence of a `task.run_function()`
        call.

       Internally, a function return is saved as a pickled file. This method unpickles each file one after
       another and yields the data. Moreover, the return data will be yielded in the same order as the functions were
       executed.

        Yields:
            Generator[object, None, None]: Generator object yielding the return data of the functions executed during
                                           task execution.
        """
        self.log.debug(f'Start generating function returns for RuntimeTask {self.name}.')

        # Ensure that the task finished its execution already
        self.join()

        for return_pkl_path in self._function_return_pkl_paths:

            if not os.path.exists(return_pkl_path):
                warnings.warn('Pickle file ' + return_pkl_path + ' with function return data does not exist. '
                              'Please check the logs.')
                yield None
                continue

            with open(return_pkl_path, 'rb') as file:
                yield pickle.load(file)

    @property
    def process(self) -> Process:
        """The process object in which the task were executed. None, if not yet or synchronously executed.
        """
        return self._process

    @property
    def env_variables(self) -> Dict:
        """Environment parameters used when executing a task.
        """
        return self._env_variables

    @env_variables.setter
    def env_variables(self, env_variables: Dict):
        """Set environment parameters used when executing a RuntimeTask.

        Args:
            env_variables: The env variables as dictionary.
        """
        self._env_variables = env_variables

    def send_file(self, local_path: str, remote_path: Optional[str] = None) -> 'RuntimeTask':
        """Create a task step for sending either a single file or a folder from localhost to another host.
        
        Args:
            local_path: Path to file on local machine.
            remote_path: Path on the remote host. Defaults to the connection working directory. See
                         `RuntimeTask.execute()` docs for further details.

        Returns:
            RuntimeTask: self.
        
        Raises:
            ValueError: If local_path is emtpy.
        """
        if not local_path:
            raise ValueError("Local path must not be empty")

        self._task_steps.append(self._TaskStep.create_send_file_instance(local_path, remote_path))

        self.log.debug(f'Step for sending the file {local_path} to {remote_path} was added to RuntimeTask {self.name}.')
        return self

    def get_file(self, remote_path: str, local_path: Optional[str] = None) -> 'RuntimeTask':
        """Create a task step for getting either a single file or a folder from another host to localhost.

        Args:
            remote_path: Path to file on host.
            local_path: Path to file on local machine. The remote file is downloaded  to the current working directory
                        (as seen by os.getcwd) using its remote filename if local_path is None. This is the default
                        behavior of fabric.

        Returns:
            RuntimeTask: self.

        Raises:
            ValueError: If remote path is emtpy.
            OSError: In case of non existent paths.
        """
        if not remote_path:
            raise ValueError("Remote path must not be empty")

        self._task_steps.append(self._TaskStep.create_get_file_instance(remote_path, local_path))

        self.log.debug(f'Step for getting the file {remote_path} to {local_path} was added to RuntimeTask {self.name}.')
        return self

    def run_command(self, command: str) -> 'RuntimeTask':
        """Create a task step for running a given shell command. 

        Args:
            command: Shell command.

        Returns:
            RuntimeTask: self.

        Raises:
            ValueError: If command is emtpy.
        """
        if not command:
            raise ValueError("Command must not be emtpy")
        self._task_steps.append(self._TaskStep.create_run_command_instance(command))

        self.log.debug(f'Step for running the command `{command}` was added to RuntimeTask {self.name}.')
        return self

    def run_function(self, function: callable, **func_kwargs) -> 'RuntimeTask':
        """Create a task step for executing a given python function on a remote host. The function will be transferred
        to the remote host via ssh and cloudpickle. The return data can be requested via the property `function_returns`

        Note:
            Hence, the function must be serializable via cloudpickle and all dependencies must be available in its
            correct versions on the remote host for now. We are planning to improve the dependency handling.

        Args:
            function: The function to be executed remotely.
            **func_kwargs: kwargs which will be passed to the function.

        Returns:
            RuntimeTask: self.
        
        Raises:
            ValueError: If function is empty.
        """

        # Attention:
        #   Please do not change the current task step composition, i.e. the order of the task steps where the current
        #   TaskStep is composed of. The reason is the order of the task log, which will break. At least, the indices
        #   in Runtime._read_info() needs to be adapted. Of course, this change might affect existing users!!!

        if not function:
            raise ValueError("Function is invalid.")

        self.log.debug(f'Start generating steps to RuntimeTask {self.name} for executing function {function.__name__}')

        steps = []

        # Create a temp directory for persisting the serialized function
        if not self._temp_dir:
            self._temp_dir = tempfile.mkdtemp()
            self.log.debug(f'Temporary directory {self._temp_dir} for RuntimeTask {self.name} on localhost created.')

        # We distinguish between local -  and remote filename because in case we execute a function in a
        # `Runtime('localhost')`, we may overwrite the files.
        input_function_name = function.__name__
        remote_file_name = self._create_file_name(input_function_name)
        local_file_name = 'local_' + remote_file_name
        local_pickle_file_path = os.path.join(self._temp_dir, local_file_name)
        remote_pickle_file_path = os.path.join(".", remote_file_name)

        local_return_file_name = 'return_' + remote_file_name
        remote_return_file_name = 'remote_' + local_return_file_name
        local_return_pickle_file_path = os.path.join(self._temp_dir, local_return_file_name)
        remote_return_pickle_file_path = os.path.join(".", remote_return_file_name)

        # Hereby we can pickle the function incl. kwargs, since we are
        # going to pickle the wrapper and not only the function itself.
        def function_wrapper():
            # Make the function return data available
            return function(**func_kwargs)

        # Create cloud pickle file of the function_wrapper() and save it locally
        with open(local_pickle_file_path, 'wb') as f:
            pickle.dump(function_wrapper, f)
            f.close()

        self.log.debug(f'Function {function.__name__} locally saved as pickle file under `{local_pickle_file_path}`.')

        # Add `TaskStep` for sending the pickled function_wrapper file to the Runtime
        steps.append(self._TaskStep.create_send_file_instance(local_pickle_file_path, remote_pickle_file_path))
        self.log.debug(f'Step for sending the file {local_pickle_file_path} to {remote_pickle_file_path} was added to '
                       f'RuntimeTask {self.name}.')

        # Ensure installation of cloudpickle
        run_cmd = 'pip install -q cloudpickle'
        steps.append(self._TaskStep.create_run_command_instance(run_cmd))
        self.log.debug(f'Step for running the command `{run_cmd}` was added to RuntimeTask {self.name}.')

        # Create commands for executing the pickled function file
        # => pickle the return data and save it as a file on the remote host so that we can get the file back to the
        # local machine. Local machine means NOT the one where `function_wrapper()` is actually executed, but
        # the one where `task.run_function()` is actually executed.
        run_function_code = 'import cloudpickle as pickle; ' \
                            'file=open(\'' + remote_pickle_file_path + '\', \'rb\'); ' \
                            'fn=pickle.load(file); file.close(); result = fn(); ' \
                            'file=open(\'' + remote_return_pickle_file_path + '\', \'wb\'); ' \
                            'pickle.dump(result, file); file.close();'

        run_cmd = 'python -c "' + run_function_code + '"'
        steps.append(self._TaskStep.create_run_command_instance(run_cmd))
        self.log.debug(f'Step for running the command `{run_cmd}` was added to RuntimeTask {self.name}.')

        # Ensure the cleanup of the created function pickle file on the Runtime
        run_cmd = 'rm ' + remote_pickle_file_path
        steps.append(self._TaskStep.create_run_command_instance(run_cmd))
        self.log.debug(f'Step for running the command `{run_cmd}` was added to RuntimeTask {self.name}.')

        # Get the return data back as pickled file
        steps.append(self._TaskStep.create_get_file_instance(remote_return_pickle_file_path,
                                                             local_return_pickle_file_path))
        self.log.debug(f'Step for getting the file {remote_return_pickle_file_path} to {local_return_pickle_file_path} '
                       f'was added to RuntimeTask {self.name}.')

        self._function_return_pkl_paths.append(local_return_pickle_file_path)

        # Cleanup the return pickle file on the remote host
        run_cmd = 'rm ' + remote_return_pickle_file_path
        steps.append(self._TaskStep.create_run_command_instance(run_cmd))
        self.log.debug(f'Step for running the command `{run_cmd}` was added to RuntimeTask {self.name}.')

        self._task_steps.append(self._TaskStep.create_run_function_instance(steps, function, func_kwargs))
        self.log.debug(f'Successfully added the final step for running the function {function.__name__} to the '
                       f'RuntimeTask {self.name}.')

        self.log.debug(f'Finished generating steps for RuntimeTask {self.name} for executing '
                       f'function {function.__name__}.')
        return self

    def execute(self, connection: Connection, debug: bool = False):
        """Execute the task on a remote host using a fabric connection.

        Note:
            Each individual task step will be executed relatively to the current directory of the fabric connection.
            Although, the current directory might have changed in the previous task step. Each directory change is
            temporary limited to a single task step.
            If the task gets executed via a `Runtime`, then the current directory will be the Runtimes working
            directory. See the `Runtime` docs for further details.
            Moreover, beside the regular Python log or the `debug` option you can access the execution logs via
            task.`execution.log`. The log gets updated after each task step.
        
        Args:
            connection: Fabric connection object managing the ssh connection to the remote host.
            debug : If `True`, stdout/stderr from the remote host will be printed to stdout. If, `False`
                    then the stdout/stderr will be written to an execution log file. Defaults to `False`.
        Raises:
            ValueError: If cxn is broken and connection can not be established.
            TaskExecutionError: If an executed task step can't be executed successfully.
            OSError: In case of file transfer and non existent paths.
        """
        exec_file_log_util = ExecutionFileLogUtil(connection.original_host, self.name)
        self._execution_log_file_path = exec_file_log_util.file_path

        # The Fabric connection will only consider the manually set working directory for connection.run()
        # but not fur connection.put() & connection.get(). Thus, we need to set the path in
        # these cases manually.
        from socket import gaierror
        try:
            working_dir = connection.run('pwd', hide=True, warn=True).stdout.replace('\n', '').replace('\r', '')
            self.log.debug(f'Current directory of connection to host {connection.original_host} is `{working_dir}`.')
        except gaierror:
            raise ValueError('Connection cannot be established. connection: ' + str(connection))

        self.log.info(f'Start executing RuntimeTask {self.name} on host {connection.original_host}')

        task_step_index = 0

        for task_step in self._task_steps:

            if task_step.type == self._TaskStep.TYPE_RUN_COMMAND:
                self._execute_run_command_step(task_step, task_step_index, connection, debug, exec_file_log_util)

            elif task_step.type == self._TaskStep.TYPE_SEND_FILE:
                self._execute_send_file(task_step, task_step_index, working_dir, connection)

            elif task_step.type == self._TaskStep.TYPE_GET_FILE:
                self._execute_get_file(task_step, task_step_index, working_dir, connection)

            if task_step.type != self._TaskStep.TYPE_RUN_FUNCTION:
                task_step_index = task_step_index + 1
                continue

            # The remote python function execution is not an elementary TaskStep (as opposed to run_command,
            # send_file, get_file) but is composed of these. Hence, we will execute the necessary steps here in order to
            # execute the desired function remotely. Those steps where generated during the RuntimeTask.run_function()
            # execution and not directly created by the user.
            if task_step.type == self._TaskStep.TYPE_RUN_FUNCTION:
                self.log.debug(f'Start executing the generated steps that are necessary to execute the python function '
                               f'`{task_step.function.__name__}` remotely.')
                for function_step in task_step.function_steps:

                    if function_step.type == self._TaskStep.TYPE_RUN_COMMAND:
                        self._execute_run_command_step(function_step, task_step_index, connection, debug, exec_file_log_util)

                    elif function_step.type == self._TaskStep.TYPE_SEND_FILE:
                        self._execute_send_file(function_step, task_step_index, working_dir, connection)

                    elif function_step.type == self._TaskStep.TYPE_GET_FILE:
                        self._execute_get_file(function_step, task_step_index, working_dir, connection)

                self.log.debug(f'Finished executing the generated steps that are necessary to execute the python '
                               f'function `{task_step.function.__name__}` remotely.')

        self.log.info(f'Finished executing RuntimeTask {self.name} on host {connection.original_host}')

    def join(self):
        """Block the execution until the `RuntimeTask` finished its asynchronous execution.

        Note:
            If self.omit_on_join is set, then the execution is omitted in order to prevent a deadlock.

        """
        if self.omit_on_join and self.process:
            self.log.debug(f'The execution of join() of RuntimeTask {self.name} is omitted, since this task is marked '
                           f'as needs explicit termination.')
            return

        if not self.process:
            return

        self.log.info(f'Start joining the process that is executing RuntimeTask {self.name}.')
        self.process.join()
        self.log.info(f'Finished joining the process that is executing RuntimeTask {self.name}.')

    def print_log(self):
        """Print the execution log. Each log entry will be printed separately. The log index will be prepended.
        """
        if not self.execution_log:
            print('The log of task ' + self.name + ' is empty!')
        else:
            print('Log of Task ' + self.name + ':')
        i = 0
        for entry in self.execution_log:
            print(str(i) + ': ' + entry)
            i += 1

    # Private helper methods & attributes
    _function_index = 0
    _FUNCTION_PREFIX = 'func'
    _PICKLE_FILENAME_EXT = 'pkl'

    @classmethod
    def _create_file_name(cls, function_name: Optional[str] = None):
        """Creates a program unique function name.
        """
        if function_name:
            file_prefix = function_name
        else:
            file_prefix = cls._FUNCTION_PREFIX
        cls._function_index += 1
        return file_prefix + str(cls._function_index) + '.' + cls._PICKLE_FILENAME_EXT

    def _execute_run_command_step(self, task_step, task_step_index: int, connection: Connection, debug: bool,
                                  exec_file_log_util: ExecutionFileLogUtil):
        self.log.debug(f'Start executing step {task_step_index} (`run_command`) from RuntimeTask {self.name} on host '
                       f'{connection.original_host}. Command: `{task_step.command}`')

        self.log.debug(f'Used environment vars on host {connection.original_host}: {self._env_variables}')

        # The output will be sent to stdout in case of debug
        # Otherwise it gets written to an execution log file
        if debug:
            result = connection.run(task_step.command, warn=True, pty=True, env=self._env_variables)
        else:
            with open(exec_file_log_util.file_path, exec_file_log_util.get_write_mode()) as log_file:
                result = connection.run(task_step.command, warn=True, pty=True, out_stream=log_file,
                                        env=self._env_variables)

        if not debug:
            self.log.debug(f'The stdout of step {task_step_index} (`run_command`) from RuntimeTask {self.name} on host '
                           f'{connection.original_host} is `{result.stdout}`.')

        self._execution_log.append(result.stdout)

        if result.exited != 0:
            raise TaskExecutionError(task_step_index, self, connection.original_host,
                                     exec_file_log_util.file_path, result.stdout)

    def _execute_send_file(self, task_step, task_step_index: int, working_dir: str, connection: Connection):
        # Ensure that we have a valid remote path
        remote_path = task_step.remote_path

        if not remote_path:
            # Manually set the working directory since fabric defaults to the os
            # user's home directory in case of file transfer
            local_dir, filename = os.path.split(task_step.local_path)
            remote_path = os.path.join(working_dir, filename)
            self.log.debug(f'Remote path was not given and manually set to the working directory {remote_path}.')

        elif remote_path.startswith('.'):
            # Relative path -> in this case it will be interpreted relative to the working directory
            remote_path = os.path.join(working_dir, remote_path[2:])
            self.log.debug(f'Remote path was relative path and changed to full path to the working directory.')

        # Update the path in the `_TaskStep` so that the actual used path is correctly stored
        task_step.remote_path = remote_path

        self.log.debug(f'Start executing TaskStep {task_step_index} (`send_file`) from RuntimeTask {self.name} '
                       f'on host {connection.original_host}. File: `local: {task_step.local_path}`, '
                       f'`remote: {task_step.remote_path}`.')

        connection.put(task_step.local_path, task_step.remote_path)

        self.log.info(f'Finished sending the file {task_step.local_path} to {task_step.remote_path} '
                      f'on host {connection.original_host}')

    def _execute_get_file(self, task_step, task_step_index: int, working_dir: str, connection: Connection):
        # Set working directory manually, so that it feels naturally for the user if he just
        # uses the filename without a path
        remote_path = task_step.remote_path
        if not remote_path.startswith('/'):
            remote_dir, filename = os.path.split(remote_path)
            remote_path = os.path.join(working_dir, filename)

        # or a relative path (-> will be interpreted relatively to the working directory)
        elif remote_path.startswith('.'):
            remote_path = os.path.join(working_dir, remote_path[2:])

        # Update the path in the `_TaskStep` so that the actual used path is correctly stored
        task_step.remote_path = remote_path

        self.log.debug(f'Start executing step {task_step_index} (`get_file`) from RuntimeTask {self.name} on '
                       f'host {connection.original_host}. File: `local: {task_step.local_path}`, '
                       f'`remote: {task_step.remote_path}`.')

        connection.get(task_step.remote_path, task_step.local_path)

        self.log.info(f'Finished transferring remote file {task_step.remote_path} from host {connection.original_host} '
                      f'to local file {task_step.local_path}.')

    class _TaskStep:
        """Represents an individual action, i.e. a `_TaskStep` within a `RuntimeTask`.
        """
        TYPE_RUN_COMMAND = 'command'
        TYPE_SEND_FILE = 'send-file'
        TYPE_GET_FILE = 'get-file'
        TYPE_RUN_FUNCTION = 'function'

        def __init__(self, step_type: str, local_path: Optional[str] = None,
                     remote_path: Optional[str] = None, command: Optional[str] = None,
                     function_steps: Optional[List['_TaskStep']] = None, function: Optional[callable] = None,
                     func_kwargs: Optional[dict] = None):
            self.type = step_type
            # Related to send_ / get_file
            self.local_path = local_path
            self.remote_path = remote_path
            # Related to run_command
            self.command = command
            # Related to run_function
            self.function = function
            self.func_kwargs = func_kwargs
            self.function_steps = function_steps

        @classmethod
        def create_get_file_instance(cls, remote_path: str, local_path: str):
            return cls(cls.TYPE_GET_FILE, local_path, remote_path)

        @classmethod
        def create_send_file_instance(cls, local_path: str, remote_path: Optional[str] = None):
            return cls(cls.TYPE_SEND_FILE, local_path, remote_path)

        @classmethod
        def create_run_command_instance(cls, command: str):
            return cls(cls.TYPE_RUN_COMMAND, command=command)

        @classmethod
        def create_run_function_instance(cls, steps: list,
                                         function: callable, func_kwargs: Optional[dict] = None):
            return cls(cls.TYPE_RUN_FUNCTION, function_steps=steps, function=function, func_kwargs=func_kwargs)

        def __repr__(self):
            return "%s(%r)" % (self.__class__, self.__dict__)


class Runtime(object):
    """A `Runtime` is the logical representation of a remote host. Typically, the host is another server or a virtual
    machine / container on another server. This python class provides several methods for utilizing remote resources
    such as the port exposure from / to a `Runtime` as well as the execution of `RuntimeTasks`. A `Runtime` has a
    working directory. Usually, the execution of a `RuntimeTask` is conducted relatively to this directory if no other
    path is explicitly given. The working directory can be manually set during the initialization. Otherwise, a
    temporary directory gets created that might eventually be removed.

    A Runtime has a working directory (property: `working_dir`) which is a temporary directory per default and gets
    deleted `atexit` in this case. If you set this directory manually, either via `__init__()` or via the property
    `working_dir` then it won't be removed. Moreover, the working directory will also be set as environment variable on
    the Runtime. It is accessible via the env variable name stated in the constant `Runtime.WORKING_DIR_ENV_VAR_NAME`.
    This might be especially of interest when executing python functions remotely.
    
    Note:
        [Passwordless ssh](https://linuxize.com/post/how-to-setup-passwordless-ssh-login/)access should be be setup in
        advance. Otherwise the connection kwargs of fabric must be used for setting up the ssh connection.

    Examples:
        ```python
        # Execute a RuntimeTask synchronously
        Runtime('host-1').execute_task(my_task, execute_async=False)
        # Expose a port from localhost to the remote host so that a service running on localhost
        # is accessible from the remote host as well.
        Runtime('host-1').expose_port_to_runtime(8786)
        # Expose a port from a remote `Runtime` to localhost so that a service running on the `Runtime`
        # is accessible from localhost as well.
        Runtime('host-1').expose_port_from_runtime(8787)
        ```
    """
    # The working directory will be set as environment variable with this env variable name
    WORKING_DIR_ENV_VAR_NAME = 'WORKING_DIR'
    LOCALHOST = 'localhost'

    # Constants used for generating process keys
    _PORT_TO_RUNTIME = '-R'
    _PORT_FROM_RUNTIME = '-L'
    _PROCESS_KEY_DELIMITER = '::'
    _TASK_PROCESS_KEY_PREFIX = 'task'

    def __init__(self, host: str, working_dir: Optional[str] = None, connection_kwargs: Optional[Dict] = None):
        """Initialization method.

        Note:
            The working directory will also be set as environment variable (see `Runtime.env_variables`) on the Runtime.
            It is accessible via the env variable name stated in the constant `Runtime.WORKING_DIR_ENV_VAR_NAME`. This
            might be especially of interest when executing functions remotely.

        Args:
            host: The host of the `Runtime`.
            working_dir: The directory which shall act as working directory. If set, then the full path will be created
                         on the remote host in case it does not exist. All individual Steps of a `RuntimeTask` will be
                         executed relatively to this directory. Defaults to None. Consequently, a temporary directory
                         will be created and used as working dir. If the working directory is a temporary one it will be
                         cleaned up either `atexit` or when calling `cleanup()` manually.

            connection_kwargs: kwargs that will be passed on to the fabric connection. Please check the fabric docs
                                 for further details.

        Raises:
            InvalidRuntimeError: If is_valid_runtime() check fails.
            PathCreationError: If the working_dir path could not be created successfully.
        """

        # Create the Logger
        self.log = logging.getLogger(__name__)

        # All attributes that need to be present before executing the valid Runtime check must be put here !!!!!
        self._host = host
        self._connection_kwargs = connection_kwargs
        self._env_variables = {}  # Will be passed on to fabric connect, so that they are available on the remote host

        # This check relies on some attributes, which must be declared above
        if not self.is_valid_runtime():
            raise InvalidRuntimeError(host)

        self._working_dir_is_temp = False  # Indicates that a temp directory acts as working directory
        self._working_dir = None
        if working_dir:
            self.working_dir = working_dir  # raises PathCreationError

        self._processes = {}  # The dict key is a generated process identifier and the value contains the process
        self._info = {}
        self._process_manager = Manager()
        self._tasks = []

        # Cleanup will be done atexit since usage of destructor may lead to exceptions
        atexit.register(self.cleanup)

        self.log.debug(f'Runtime {self.host} initialized.')

    def __repr__(self):
        return "%s(%r)" % (self.__class__, self.__dict__)

    def __str__(self):
        return self.class_name + ': ' + self.host

    @property
    def working_dir(self) -> str:
        """The path of the working directory that was set during object initialization.

        Note:
            The working directory will also be set as environment variable on the Runtime. It is accessible via the
            env variable name stated in the constant `Runtime.WORKING_DIR_ENV_VAR_NAME`. This might be especially of
            interest when executing python functions remotely.
            Moreover, The full path will be created on the remote host in case it does not exist.

        Returns:
            str: The path of the working directory.
        """
        self._create_working_dir_if_not_exists()
        return self._working_dir

    @working_dir.setter
    def working_dir(self, working_dir: str):
        """ Setter of the working directory. This will also update the related env variable.

        Note:
            The full path will be created on the remote host in case it does not exist.

        Args:
            working_dir: The full path to the working directory.

        Raises:
            PathCreationError: If the working_dir path could not be created successfully.
        """
        self.create_dir(working_dir)  # raises PathCreationError
        self._working_dir = working_dir
        self._working_dir_is_temp = False
        self._env_variables.update({self.WORKING_DIR_ENV_VAR_NAME: self._working_dir})
        self.log.debug(f'The working directory  `{self._working_dir}` of Runtime {self.host} was set as environment '
                       f'variable with the name {self.WORKING_DIR_ENV_VAR_NAME}.')

    @property
    def task_processes(self) -> List[Process]:
        """All processes that were started to execute a `RuntimeTask` asynchronously.

        Returns:
            List[Process]: RuntimeTask processes.
        """
        processes = []
        for process_key, process in self._processes.items():
            if self.is_runtime_task_process(process_key):
                processes.append(process)
        return processes

    @property
    def function_returns(self) -> Generator[object, None, None]:
        """The return data produced by Python functions which were executed as a consequence of
           `task.run_function()`. The call will be passed on to the `function_returns` property of the `RuntimeTask`.
           The order is determined by the order in which the `RuntimeTasks` were executed in the `Runtime`.

        Yields:
            Generator[object, None, None]: Generator object yielding the return data of the functions executed during
                                           task execution.
        """
        self.log.debug(f'Start generating function returns for Runtime {self.host}.')
        for task in self._tasks:
            for function_return in task.function_returns:
                yield function_return

    @property
    def host(self) -> str:
        """The host of the runtime.

        Returns:
            str:  The host of the runtime.
        """
        return self._host

    @property
    def info(self) -> dict:
        """Information about the runtime.

        Returns:
            dict: Runtime information.
        """
        if not self._info:
            self._info = self._read_info()
        return self._info

    @property
    def os(self) -> str:
        """Operating system information.

        Returns:
            str: OS.
        """
        if not self._info:
            self._info = self._read_info()
        return self._info['os']

    @property
    def cpu_cores(self) -> int:
        """Information about the available CPUs. If you are in a container
        the CPU quota will be given if set. Otherwise, the number of physical CPUs
        on the host machine is given.

        Returns:
            str: CPU quota.
        """
        if not self._info:
            self._info = self._read_info()
        return int(self._info['cpu_cores'])

    @property
    def memory(self) -> int:
        """Information about the total memory in bytes.

        Returns:
            str: Total memory in bytes.
        """
        if not self._info:
            self._info = self._read_info()
        return int(self._info['memory'])

    @property
    def memory_in_mb(self) -> int:
        """Memory information in mb.

        Returns:
            int: Total memory in mega bytes.
        """
        return int(self.memory / 1024 / 1024)

    @property
    def python_version(self) -> str:
        """The installed python version.

        Returns:
            str: Python version.
        """
        if not self._info:
            self._info = self._read_info()
        return self._info['python_version']

    @property
    def gpus(self) -> list:
        """GPU information as list. Each list entry contains information for one GPU.

        Returns:
            list: List with GPU information.
        """
        if not self._info:
            self._info = self._read_info()
        return self._info['gpus']

    @property
    def gpu_count(self) -> int:
        """The count of GPUs.

        Returns:
            int: The number of GPUs
        """
        return len(self.gpus)

    @property
    def class_name(self) -> str:
        """The class name  as string. """
        return self.__class__.__name__

    @property
    def alive_process_count(self) -> int:
        """The number of alive processes.

        Returns:
            int: The count.
        """
        return len(self.get_processes(only_alive=True))

    @property
    def alive_task_process_count(self) -> int:
        """The number of alive processes which were started to execute a `RuntimeTask`.

        Returns:
            int: The count.
        """
        return len(self.task_processes)

    @property
    def env_variables(self) -> Dict[str, str]:
        """The environment variables for the Runtime. These variables are accessible on the Runtime and can be used
        when executing Python functions or shell commands.

        Note:
            The working directory is always accessible as environment variable on the Runtime. The respective variable
            name is given by the value of the constant `self.WORKING_DIR_ENV_VAR_NAME`.
        """
        return self._env_variables

    @env_variables.setter
    def env_variables(self, env_variables: Dict[str, str]):
        """Setter for the environment variables.

        Args:
            env_variables: The new env var dictionary.
        """
        self._env_variables = env_variables
        if self.WORKING_DIR_ENV_VAR_NAME not in self._env_variables:
            self._env_variables.update({self.WORKING_DIR_ENV_VAR_NAME: self._working_dir})

    def add_env_variables(self, env_variables: Dict):
        """Update the environment variables. If a variable already exists it gets updated and if not it will be added.

        Args:
            env_variables: The env variables used for the update.
        """
        self._env_variables.update(env_variables)

    @classmethod
    def is_runtime_task_process(cls, process_key: str) -> bool:
        """Checks if the process which belongs to a given `process_key` was started to execute a `RuntimeTask` based on
        an internal naming scheme of the process keys.

        Args:
            process_key: The generated process identifier.
        Returns:
            bool: True, if process was started to execute a `RuntimeTask`
        """
        key_splits = process_key.split(cls._PROCESS_KEY_DELIMITER)
        return True if key_splits[1] == cls._TASK_PROCESS_KEY_PREFIX else False

    @classmethod
    def is_port_exposure_process(cls, process_key: str) -> bool:
        """Check if the process which belongs to the given `process_key` is used for exposing a port, i.e. keeping
        an ssh tunnel alive.

        Args:
            process_key (str): The generated process identifier.
        Returns:
            bool: True, if process is used for port exposure.
        """
        key_splits = process_key.split(cls._PROCESS_KEY_DELIMITER)
        return True if key_splits[1] == cls._PORT_FROM_RUNTIME or key_splits[1] == cls._PORT_TO_RUNTIME else False

    def is_valid_runtime(self) -> bool:
        """Checks if a given host is a valid `Runtime`.

        Returns:
            bool: True, if it is a valid remote runtime.
        """
        self.log.debug(f'Start executing `is_valid_runtime()` on Runtime {self.host}')

        # Paramiko is only used by fabric and thus not needed in our project requirements
        from paramiko.ssh_exception import SSHException, NoValidConnectionsError

        try:
            # use a relatively high timeout to prevent errors when sshing with slow network connections
            cxn = Connection(host=self.host, connect_kwargs=self._connection_kwargs,
                             inline_ssh_env=True, connect_timeout=2)
            stdout = cxn.run('python --version', env=self._env_variables, warn=False, hide=True, pty=True, timeout=10).stdout
        except NoValidConnectionsError:
            self.log.debug(f'No valid ssh connection to host {self.host}.')
            return False
        except SSHException:
            self.log.debug(f'connection.run() threw a SSHException during is_valid_runtime of host {self.host}.')
            return False
        except ValueError:
            self.log.debug(f'connection.run() threw a ValueError during is_valid_runtime of host {self.host}.')
            return False

        stdout = stdout.replace('\n', '').replace('\r', '')
        self.log.debug(f'stdout of python version check on host {self.host} is `{stdout}`.')

        if not stdout:
            return False

        # Example: `Python 3.6.8 :: Anaconda, Inc.`
        python_version = stdout.split()[1].split('.')

        if not python_version:
            return False
        elif int(python_version[0]) > 3:  # Stay future-proof?
            self.log.warning('The lib was originally created for Python 3.6 and is not yet tested for '
                             'Python >= Python 4')
            return True
        elif int(python_version[0]) == 3 and int(python_version[1]) >= 6:
            return True
        else:
            return False

    def get_free_port(self, ports: List[int]) -> int:
        """Returns the first port from the list which is currently not in use in the `Runtime`.

        Args:
             ports: The list of ports that will be used to check if the port is currently in use.

        Returns:
            int: The first port from the list which is not yet used within the whole group.

        Raises:
            NoPortsLeftError: If the port list is empty and no free port was found yet.
        """
        for port in ports:
            if self.has_free_port(port):
                return port

        raise NoPortsLeftError()

    def execute_task(self, task: RuntimeTask, execute_async: Optional[bool] = True,
                     omit_on_join: bool = False, debug: bool = False):
        """Execute a given `RuntimeTask` in the `Runtime`.

        Note:
            Each execution will initialize the execution log of the `RuntimeTask`.

        Args:
            task: The RuntimeTask to be executed.
            execute_async: The execution will be done in a separate process if True. Defaults to True.
            omit_on_join: If True, then a call to join() won't wait for the termination of the corresponding process.
                          Defaults to False. This parameter has no effect in case of synchronous execution.
            debug : If `True`, stdout/stderr from the remote host will be printed to stdout. If, `False`
                    then the stdout/stderr will be written to execution log files. Defaults to `False`.

        Raises:
            TaskExecutionError: If an executed task step can't be executed successfully.
        """
        self._create_working_dir_if_not_exists()
        async_str = ' asynchronously ' if execute_async else ' synchronously '
        self.log.info(f'Start executing task {task.name} {async_str} on {self.host}')

        # Pass env parameters on to the RuntimeTask
        task.env_variables = self._env_variables

        # Wrapper needed to ensure execution from the Runtime's working directory
        def execute_remote_wrapper():
            cxn = self._fabric_connection
            with cxn.cd(self._working_dir):
                task.execute(cxn, debug=debug)

        if execute_async:
            # Initialize logs with managed list (multiprocessing)
            # => We can access the logs although it is executed in another process
            task._execution_log = self._process_manager.list()
            task.omit_on_join = omit_on_join
            process = Process(target=execute_remote_wrapper)
            process.start()
            self._processes.update({self._create_process_key_for_task_execution(task): process})
            task._process = process
        else:
            task._execution_log = []
            execute_remote_wrapper()

        self._tasks.append(task)

    def send_file(self, local_path: str, remote_path: Optional[str] = None, execute_async: Optional[bool] = False) \
            -> 'RuntimeTask':
        """Send either a single file or a folder from the manager to the Runtime.

        Note:
            This method is a convenient wrapper around the RuntimeTask's send file functionality. But it directly
            executes the file transfer in contrast to the send_file() method of the RuntimeTask.

        Args:
            local_path: Path to file on local machine.
            remote_path: Path on the Runtime. Defaults to the self.working_dir. See
                         `RuntimeTask.execute()` docs for further details.
            execute_async: The execution will be done in a separate process if True. Defaults to False.

        Returns:
            RuntimeTask: The task that were internally created for the file transfer.

        Raises:
            ValueError: If local_path is emtpy.
            TaskExecutionError: If an executed task step can't be executed successfully.
            OSError: In case of non existent paths.e
        """
        async_str = ' asynchronously ' if execute_async else ' synchronously '
        self.log.debug(f'Start sending local file `{local_path}` to Runtime {self.host} {async_str}. Given remote path:'
                       f' {remote_path}`.')
        task = RuntimeTask(f'send-file-{local_path}-to-{self.host}')
        task.send_file(local_path, remote_path)
        self.execute_task(task, execute_async)
        return task

    def get_file(self, remote_path: str, local_path: Optional[str] = None, execute_async: Optional[bool] = False) -> 'RuntimeTask':
        """Get either a single file or a folder from the Runtime to the manager.

        Note:
            This method is a convenient wrapper around the RuntimeTask's get file functionality. But it directly
            executes the file transfer in contrast to the get_file() method of the RuntimeTask.

        Args:
            remote_path: Path to file on host.
            local_path: Path to file on local machine (i.e. manager). The remote file is downloaded  to the current
                        working directory (as seen by os.getcwd) using its remote filename if local_path is None.
                        This is the default behavior of fabric.Connection.get().
            execute_async: The execution will be done in a separate process if True. Defaults to False.

        Returns:
            RuntimeTask: self.

        Raises:
            ValueError: If remote path is emtpy.
        """
        async_str = ' asynchronously ' if execute_async else ' synchronously '
        self.log.debug(f'Start transferring the file `{remote_path}` of Runtime {self.host} {async_str} to local path:'
                       f' {local_path}`.')
        task = RuntimeTask(f'get-file-{remote_path}-from-{self.host}')
        task.get_file(remote_path, local_path)
        self.execute_task(task, execute_async)
        return task

    def execute_function(self, function: callable, execute_async: bool = False, debug: bool = False, **func_kwargs) -> 'RuntimeTask':
        """Execute a Python function on the Runtime.

        Note:
            Internally, creates a RuntimeTask for executing the given python function on a remote host. The function
            will be transferred to the remote host via ssh and cloudpickle. The return data can be requested via the
            property `function_returns` of the Runtime or of the returned RuntimeTask. Hence, the function must be
            serializable via cloudpickle and all dependencies must be available in its correct versions on the Runtime.

        Args:
            function: The function to be executed remotely.
            execute_async: The execution will be done in a separate process if True. Defaults to False.
            debug : If `True`, stdout/stderr from the remote host will be printed to stdout. If, `False`
                    then the stdout/stderr will be written to execution log files. Defaults to `False`.
            **func_kwargs: kwargs which will be passed to the function.

        Returns:
            RuntimeTask: self.

        Raises:
            ValueError: If function is empty.
            TaskExecutionError: If there was an error during the execution.
        """
        task = RuntimeTask(f'execute-function-{function.__name__}').run_function(function, **func_kwargs)
        self.execute_task(task, execute_async, debug)
        return task

    def _create_working_dir_if_not_exists(self):
        if not self._working_dir:
            working_dir = self.create_tempdir()
            self._working_dir_is_temp = True
            self.log.debug(f'Temporary directory {working_dir} created as working directory on Runtime '
                           f'{self._host}')
            self.working_dir = working_dir  # This will call the setter of self._working_dir

    def print_log(self):
        """Print the execution logs of each `RuntimeTask` that was executed in the `Runtime`.
        """
        for task in self._tasks:
            task.print_log()

    def clear_tasks(self):
        """Clears all internal state related to `RuntimeTasks`.
        """
        self.log.info(f'Clear all RuntimeTasks and kill related processes on Runtime {self.host}')
        self._tasks = []
        self._processes = {key: value for key, value in self._processes.items()
                           if not Runtime.is_runtime_task_process(key)}

    def expose_port_to_runtime(self, local_port: int, runtime_port: Optional[int] = None) -> str:
        """Expose a port from localhost to the `Runtime` so that all traffic on the `runtime_port` is forwarded to the
        `local_port` on localhost.
        
        Args:
            local_port: The port on the local machine.
            runtime_port: The port on the runtime. Defaults to `local_port`.

        Returns:
            str: Process key, which can be used for manually stopping the process running the port exposure for example.
        """
        if not runtime_port:
            runtime_port = local_port
        if local_port == runtime_port and self._host == self.LOCALHOST:
            # We can not forward a port to itself!
            self.log.debug(f'Port ({str(local_port)}) exposure skipped on Runtime {self.host}. Can\'t expose a port to '
                           f'myself.')
            return ''
        elif not self.has_free_port(runtime_port):
            raise PortInUseError(runtime_port, runtime=self)

        proc = Process(target=self._forward_runtime_port_to_local,
                       kwargs={'local_port': local_port, 'runtime_port': runtime_port})
        proc.start()
        key = self._create_process_key_for_port_exposure(self._PORT_TO_RUNTIME, local_port, runtime_port)
        self._processes.update({key: proc})
        time.sleep(0.1)  # Necessary to prevent collisions with MaxStartup restrictions
        self.log.info(f'Local port {str(local_port)} exposed to Runtime {self._host} on port {str(runtime_port)}')
        return key

    def expose_port_from_runtime(self, runtime_port: int, local_port: Optional[int] = None) -> str:
        """Expose a port from a `Runtime` to localhost so that all traffic to the `local_port` is forwarded to the
        `runtime_port` of the `Runtime`. This corresponds to local port forwarding in ssh tunneling terms.

        Args:
            runtime_port: The port on the runtime.
            local_port: The port on the local machine. Defaults to `runtime_port`.

        Returns:
            str: Process key, which can be used for manually stopping the process running the port exposure.
        """
        if not local_port:
            local_port = runtime_port
        if local_port == runtime_port and self._host == self.LOCALHOST:
            # We can not forward a port to itself!
            self.log.debug(f'Port ({str(local_port)}) exposure skipped on Runtime {self.host}. Can\'t expose a port to '
                           f'myself.')
            return ''
        elif not _utils.localhost_has_free_port(local_port):
            raise PortInUseError(runtime_port, runtime=self)

        proc = Process(target=self._forward_local_port_to_runtime,
                       kwargs={'local_port': local_port, 'runtime_port': runtime_port})
        proc.start()
        key = self._create_process_key_for_port_exposure(self._PORT_FROM_RUNTIME, local_port, runtime_port)
        self._processes.update({key: proc})
        time.sleep(0.1)  # Necessary to prevent collisions with MaxStartup restrictions
        self.log.info(f'Port {str(runtime_port)} from runtime {self._host} exposed to local port {str(local_port)}')
        return key

    def get_process(self, key: str) -> Process:
        """Get an individual process by process key.
        
        Args:
            key: The key identifying the process.
            
        Returns:
            Process: The desired process.

        Raises:
            ValueError: Unknown process key.
        """
        if key not in self._processes:
            raise ValueError('Unknown process key.')
        return self._processes[key]

    def get_processes(self, only_alive: bool = False) -> Dict[str, Process]:
        """Get all managed processes or only the alive ones as dictionary with the process key as dict key. An
        individual process can be retrieved by key via `get_process()`.
        
        Args:
            only_alive: True, if only alive processes shall be returned instead of all. Defaults to False.
            
        Returns:
            Dict: Dictionary with process keys as dict keys and the respective processes as dict values.

        """
        if only_alive:

            ret_processes = {}

            for key, process in self._processes.items():
                if process.is_alive():
                    ret_processes.update({key: process})
            return ret_processes

        else:
            return self._processes

    def stop_process(self, key: str):
        """Stop a process by its key. 

        Args:
            key: The key identifying the process.
        
        Raises:
            ValueError: Unknown process key.
        """
        if key not in self._processes:
            raise ValueError('Unknown process key.')
        self._processes[key].terminate()
        self.log.debug(f'Process with key {key} stopped in Runtime {self.host}')

    def has_free_port(self, port: int) -> bool:
        """Checks if the port is available on the runtime. 
        
        Args:
            port: The port which will be checked.

        Returns:
            bool: True if port is free, else False.
        """
        self.log.debug(f'Checking if port {str(port)} is free on Runtime {self.host}')

        with self._fabric_connection as cxn:
            cmd_str = 'python -c "import socket;print(\'free\') ' \
                      'if socket.socket(socket.AF_INET, socket.SOCK_STREAM).connect_ex((\'localhost\', ' + str(port) + \
                      ')) else None"'
            res = cxn.run(cmd_str, hide=True)
            return True if res.stdout else False

    def print_info(self):
        """Print the Runtime info formatted as table.
        """
        info = self.info
        print('\u001b[1mInformation of `' + self.class_name + '` ' + self.host + ':\u001b[0m')
        for key, value in info.items():

            if key == 'memory':
                display_value = str(self.memory_in_mb) + ' mb'
            elif isinstance(value, list):
                display_value = ''
                for gpu in value:
                    display_value = '{}'.format(gpu)
            else:
                display_value = value

            print('{:<8} {:<8}'.format(key, display_value))

    def check_filter(self,
                     gpu_required: bool = False,
                     min_memory: Optional[int] = None,
                     min_cpu_cores: Optional[int] = None,
                     installed_executables: Union[str, List[str], None] = None,
                     filter_commands: Union[str, List[str], None] = None) -> bool:
        """Checks the `Runtime` object for certain filter criteria.

        Args:
            gpu_required: True, if gpu availability is required. Defaults to False.
            min_memory: The minimal amount of memory in MB. Defaults to None, i.e. not restricted.
            min_cpu_cores: The minimum number of cpu cores required. Defaults to None, i.e. not restricted.
            installed_executables: Possibility to check if an executable is installed. E.g. if the executable `ping` is
                                   installed.
            filter_commands: Shell commands that can be used for generic filtering. See examples. A filter command must
                             echo true to be evaluated to True, everything else will be interpreted as False. Defaults
                             to None.

        Returns:
            bool: True, if all filters were successfully checked otherwise False.

        Examples:
            ```python
            # Check if the `Runtime` has a specific executable installed
            # such as `ping` the network administration software utility.
            check_passed = runtime.check_filter(installed_executables='ping')
            # Check if a variable `WORKSPACE_VERSION` is set on the `Runtime`
            filter_str = '[ ! -z "$WORKSPACE_VERSION" ] && echo "true" || echo "false"'
            check_passed = runtime.check_filter(filer_commands=filter_str)
            ```
        """
        self.log.debug(f'Start executing check_filter() for Runtime {self.host}')

        all_filters_checked = True

        if gpu_required and not self.gpus:
            self.log.debug(f'Runtime  {self.host} does not have GPUs.')
            all_filters_checked = False

        if min_memory and self.memory_in_mb < min_memory:
            self.log.debug(f'Runtime {self.host} has only {str(self.memory_in_mb)} mb instead of {str(min_memory)} as required.')
            all_filters_checked = False

        if min_cpu_cores and self.cpu_cores < min_cpu_cores:
            self.log.debug(f'Runtime {self.host} has only {str(self.cpu_cores)} instead of {str(min_cpu_cores)} as required.')
            all_filters_checked = False

        if installed_executables:
            for executable_name in _utils.create_list_from_parameter_value(installed_executables):
                if not self._has_executable_installed(str(executable_name)):
                    self.log.debug(f'Runtime {self.host} does not have executable {str(executable_name)} installed.')
                    all_filters_checked = False

        if filter_commands:
            for filter_command in _utils.create_list_from_parameter_value(filter_commands):
                if not self._filter_command_checked(str(filter_command)):
                    self.log.debug(f'Filter filter_commands could not be checked successfully on Runtime' f' {self.host}.')
                    all_filters_checked = False

        return all_filters_checked

    def create_tempdir(self) -> str:
        """Create a temporary directory and return its name/path.

        Returns:
            str: The name/path of the directory.
        """
        with self._fabric_connection as cxn:
            cmd_str = 'python -c "import tempfile; print(tempfile.mkdtemp())"'
            res = cxn.run(cmd_str, hide=True)
            path = res.stdout.split("\n")[0]
            if not path:
                path = res.stdout
            self.log.debug(f'Temporary directory {path} created on Runtime {self.host}')
            return path

    def create_dir(self, path: str):
        """Create a directory. All folders in the path will be created if not existing.

        Args:
            path: The full path of the directory to be created.

        Raises:
            PathCreationError: If the path could not be created successfully.
        """
        try:
            with self._fabric_connection as cxn:
                cmd_str = 'mkdir -p ' + path
                res = cxn.run(cmd_str, hide=True)
                if res.stderr:
                    raise PathCreationError(path, self.host)
                else:
                    self.log.debug(f'Directory {path} created on Runtime {self.host}')
        except Exception:
            raise PathCreationError(path, self.host)

    def delete_dir(self, path: str) -> bool:
        """Delete a directory recursively. If at least one contained file could not be removed then False is returned.

        Args:
            path: The full path of the directory to be deleted.

        Returns:
            bool: True if the directory could be deleted successfully.
        """
        from invoke.exceptions import ThreadException
        with self._fabric_connection as cxn:
            cmd_str = f'rm -r {path} 2> /dev/null'

            try:
                res = cxn.run(cmd_str, warn=True)
            except ThreadException:
                self.log.warning(f'ThreadException occured when deleting the directory {path}')
                return True

            if res.ok:
                self.log.debug(f'Directory {path} deleted from Runtime {self.host}')
                return True
            else:
                self.log.debug(f'Directory {path} may not be deleted on Runtime {self.host}')
                return False

    def join(self):
        """Blocks until `RuntimeTasks` which were started via the `runtime.execute_task()` method terminated.
        """
        self.log.info(f'Joining all processes executing a RuntimeTask that were started via the Runtime {self.host}')
        for task in self._tasks:
            task.join()

    def cleanup(self):
        """Release all acquired resources and terminate all processes.
        """
        self.log.info(f'Start cleanup of Runtime {self.host}.')
        for key, process in self._processes.items():
            process.terminate()
            process.join()
            if process.is_alive():
                self.log.warning(f'Process with key {key} could not be terminated')
            else:
                self.log.debug(f'Process with key {key} terminated')
        if self._working_dir_is_temp and self._working_dir:
            success = self.delete_dir(self._working_dir)
            if success:
                self.log.debug(f'Temporary directory {self.working_dir} of Runtime {self.host} removed.')
                self._working_dir = None
                self._working_dir_is_temp = False
            else:
                self.log.warning(f'Temporary directory {self.working_dir} of Runtime {self.host} could not be'
                                 f' removed.')

        for task in self._tasks:
            task.cleanup()

    def echo(self, msg: str) -> str:
        """Convenient method for echoing a string on the `Runtime` and returning the result.
        """
        return self._fabric_connection.run(f'echo {msg}', env=self._env_variables, hide=True).stdout

    # - Private methods -#

    def _create_process_key_for_port_exposure(self, direction: str, local_port: int, runtime_port: int) -> str:
        """Create a process key for processes exposing ports, i.e. keeping ssh tunnels open.
        This key will act as an identifier for internally generated processes.
        
        Raises:
            ValueError: If direction has an invalid value.
        """
        if not local_port:
            local_port = runtime_port
        if not runtime_port:
            runtime_port = local_port

        delimiter = self._PROCESS_KEY_DELIMITER

        if direction == self._PORT_FROM_RUNTIME:
            return self.host + delimiter + self._PORT_FROM_RUNTIME + delimiter + str(local_port) + delimiter + \
                   str(runtime_port)
        elif direction == self._PORT_TO_RUNTIME:
            return self.host + delimiter + self._PORT_TO_RUNTIME + delimiter + str(runtime_port) + delimiter + \
                   str(local_port)
        else:
            raise ValueError(direction + ' is not a supported runtime process prefix type')

    def _create_process_key_for_task_execution(self, task: RuntimeTask) -> str:
        """Create a process key for processes started to execute a `RuntimeTasks` asynchronously
        This key will act as an identifier for internally generated processes.
        """
        return self.host + self._PROCESS_KEY_DELIMITER + self._TASK_PROCESS_KEY_PREFIX + self._PROCESS_KEY_DELIMITER + \
               task.name

    @classmethod
    def _create_executable_installed_shell_cmd(cls, executable: str) -> str:
        return 'hash ' + executable + ' 2>/dev/null && echo "true" || echo ""'

    def _has_executable_installed(self, executable_name: str) -> bool:
        """Checks if an executable is installed on the runtime.
        """
        shell_cmd = self._create_executable_installed_shell_cmd(executable_name)
        return self._filter_command_checked(shell_cmd)

    def _filter_command_checked(self, shell_cmd: str) -> bool:
        task = RuntimeTask('_filter_command_checked')
        task.run_command(shell_cmd)
        self.execute_task(task, execute_async=False)

        # Check the last log entry for the string true
        result = str(task.execution_log[len(task.execution_log) - 1])
        return True if result.lower() == 'true' else False

    @property
    def _fabric_connection(self) -> Connection:
        """Get a new fabric connection to the runtime.

        Note: We set the `fabric.Connection` parameter `inline_ssh_env=True`.

        Raises:
            ValueError: If user or port values are given via both host shorthand
                        and their own arguments.
        """
        from socket import gaierror
        try:
            self.log.debug(f'Create new fabric connection to host {self.host} via _fabric_connection.')
            return Connection(host=self.host, connect_kwargs=self._connection_kwargs, inline_ssh_env=True)
        except gaierror:
            raise ValueError('Cannot establish SSH connection to host ' + self.host)

    def _forward_local_port_to_runtime(self, local_port: int, runtime_port: Optional[int] = None):
        """Creates ssh connection to the runtime and creates then a ssh tunnel
        from `localhost`:`local_port to `runtime`:`runtime_port`.
        """
        if not runtime_port:
            runtime_port = local_port

        with self._fabric_connection.forward_local(local_port, runtime_port):
            while True:
                time.sleep(1000)

    def _forward_runtime_port_to_local(self, runtime_port: int, local_port: Optional[int] = None):
        """Creates ssh connection to the runtime and then creates a ssh tunnel
        from `runtime`:`runtime_port` to `localhost`:`local_port`.
        """
        if not local_port:
            local_port = runtime_port

        with self._fabric_connection.forward_remote(runtime_port, local_port):
            while True:
                time.sleep(1000)

    def _read_info(self) -> dict:
        """Read the host machine information.
        """
        self.log.debug(f'Read Runtime information of Runtime {self.host}')
        return _utils.read_host_info(self.host)


"""Module comprising the abstract RuntimeCluster class with its related `launcher strategy` classes.

Note: The design of the launcher classes follows the strategy pattern.
"""
from typing import List, Optional, Dict, Union

import atexit
import logging

from subprocess import Popen
from lazycluster.runtime_mgmt import RuntimeGroup


class MasterLauncher(object):
    """Abstract class for implementing the strategy for launching the master instance of the cluster.
    """

    def __init__(self, runtime_group: RuntimeGroup):
        """Initialization method.

        Args:
            runtime_group: The group where the workers will be started.
        """
        # Create the Logger
        self.log = logging.getLogger(__name__)

        self._group = runtime_group
        self._port = None            # Needs to be set in self.start()
        self._process: Optional[Popen] = None  # Needs to be set in self.start()

        self.log.debug('MasterLauncher initialized.')

    @property
    def port(self) -> int:
        """The port where the master instance is started on. Will be None if not yet started.

        Returns:
            int: The master port.
        """
        return self._port

    @property
    def process(self) -> Popen:
        """The process object where the master instance was started in.

        Returns:
             Popen: The process object.
        """
        return self._process

    def start(self, ports: Union[List[int], int], timeout: int = 3, debug: bool = False) -> List[int]:
        """Launch a master instance.

        Note:
            If you create a custom subclass of MasterLauncher which will not start the master instance on localhost
            then you should pass the debug flag on to `execute_task()` of the `RuntimeGroup` or `Runtime` so that you
            can benefit from the debug feature of `RuntimeTask.execute()`.

        Args:
            ports: Port where the master should be started. If a list is given then the first port that is free in the
                  `RuntimeGroup` will be used. The actual chosen port can requested via the property `port`.
            timeout: Timeout (s) after which an MasterStartError is raised if master instance not started yet.
            debug: If `True`, stdout/stderr from the runtime will be printed to stdout of localhost. If, `False` then
                   the stdout/stderr will be added to python logger with level debug after each task step. Defaults to
                   `False`.

        Returns:
            List[int]: In case a port list was given the updated port list will be returned. Otherwise an empty list.

        Raises:
            PortInUseError: If a single port is given and it is not free in the `RuntimeGroup`.
            NoPortsLeftError: If a port list was given and none of the ports is actually free in the `RuntimeGroup`.
            MasterStartError: If master was not started after the specified `timeout`.
        """

        # The actual values for the following instance attributes need to be set within the concrete implementations:
        # - self._port
        # - self._process (In case the master is started in a process managed by the MasterLauncher instance )

        raise NotImplementedError

    def cleanup(self):
        """Release all resources.
        """
        self.log.debug('Cleaning up MasterLauncher ...')
        if self._process:
            self.process.terminate()


class WorkerLauncher(object):
    """Abstract class for implementing the strategy for launching worker instances within a RuntimeGroup.

    In order to implement a new concrete `WorkerLauncher` subclass you need to implement the start method. Please
    consider the comments of the start method because some internal variables need to be set in the concrete
    implementation.

    Moreover, the `setup_worker_ssh_tunnels()` method can be used to setup ssh tunnels so that all entities can talk to
    each other.
    """

    def __init__(self, runtime_group: RuntimeGroup):
        """Initialization method.

        Args:
            runtime_group: The group where the workers will be started in.
        """
        # Create the Logger
        self.log = logging.getLogger(__name__)

        self._group = runtime_group
        self._ports_per_host: Dict[str, List[int]] = {}  # Needs to be set in `self.start()`

        self.log.debug('Worker launcher initialized')

    def __repr__(self):
        return "%s(%r)" % (self.__class__, self.__dict__)

    @property
    def ports_per_host(self) -> Dict[str, List[int]]:
        """Dictionary with the host as key and a port list as value. The list contains all ports where a worker instance
        is reachable on the respective host.

        Returns:
            Dict[str, List[int]]: The ports per host as a dictionary.
        """
        return self._ports_per_host

    def start(self, worker_count: int, master_port: int, ports: List[int], debug: bool = False) -> List[int]:
        """Launches the worker instances in the `RuntimeGroup`.

        Args:
            worker_count: The number of worker instances to be started in the group.
            master_port:  The port of the master instance.
            ports: The ports to be used for starting the workers. Only ports from the list will be chosen that are
                   actually free.
            debug: If `True`, stdout/stderr from the runtime will be printed to stdout of localhost. If, `False` then
                   the stdout/stderr will be added to python logger with level debug after each task step. Defaults to
                   `False`.

        Returns:
            List[int]: The updated port list after starting the workers, i.e. the used ones were removed.

        Raises:
            NoPortsLeftError: If there are not enough free ports for starting all workers.
        """

        # 1. The actual values for the following instance attributes need to be set within the concrete implementations:
        # - self._ports_per_host

        # 2. Raise `NoPortsLeftError` if there were not sufficient free ports in the list

        raise NotImplementedError

    def setup_worker_ssh_tunnels(self):
        """Set up ssh tunnel for workers such that all communication is routed over the
        local machine and all entities can talk to each other on localhost.

        Note:
            This method needs to be called if the communication between the worker instances is necessary, e.g. in case
            of DASK or Apache Flink, where data needs to be shuffled between the different entities.

        Raises:
            ValueError: If host is not contained.
            PortInUseError: If `group_port` is occupied on the local machine.
            NoPortsLeftError: If `group_ports` was given and none of the ports was free.
        """
        self.log.info('Setting up ssh tunnel for inter worker communication.')
        for host, ports in self.ports_per_host.items():
            for worker_port in ports:
                self._group.expose_port_from_runtime_to_group(host, worker_port)  # Raises all errors

    def cleanup(self):
        """Release all resources.
        """
        self.log.debug('Cleaning up WorkerLauncher ...')
        self.log.debug('No WorkerLauncher resources to be released')


class RuntimeCluster(object):
    """Abstract cluster class.

    All further cluster implementations should inherit from this class either directly (e.g. the abstract class
    `MasterWorkerCluster`) or indirectly (e.g. the DaskCluster which is an concrete implementation of the
    `MasterWorkerCluster`).
    """

    def __repr__(self):
        return "%s(%r)" % (self.__class__, self.__dict__)


class MasterWorkerCluster(RuntimeCluster):
    """Class for clusters following a master-worker architecture.

    Usually you want to inherit from this class and do not want to use it directly. It is recommended to treat this
    class as an abstract class or an interface.

    Examples:
        Create a cluster with all `Runtimes` detected by the `RuntimeManager`.
        ´´´python
        from lazycluster import RuntimeManager
        cluster = MyMasterWorkerClusterImpl(RuntimeManager().create_group())
        cluster.start()
        ´´´
        Use different strategies for launching the master and the worker instance as the default ones by providing
        custom implementation of `MasterLauncher` and `WorkerLauncher`.
        ´´´python
        cluster = MyMasterWorkerClusterImpl(RuntimeManager().create_group(),
                                            MyMasterLauncherImpl(),
                                            MyWorkerLauncherImpl)
        cluster.start()
        ```
    """

    DEFAULT_MASTER_PORT = 60000
    DEFAULT_PORT_RANGE_START = 60001  # Can be overwritten in subclasses
    DEFAULT_PORT_RANGE_END = 60200    # Can be overwritten in subclasses

    def __init__(self, runtime_group: RuntimeGroup, ports: Optional[List[int]] = None,
                 master_launcher: Optional[MasterLauncher] = None,
                 worker_launcher: Optional[WorkerLauncher] = None):
        """Initialization method.

        Args:
            runtime_group: The `RuntimeGroup` contains all `Runtimes` which can be used for starting the cluster
                           entities.
            ports: The list of ports which will be used to instantiate a cluster. Defaults to
                   list(range(self.DEFAULT_PORT_RANGE_START, self.DEFAULT_PORT_RANGE_END).)
            master_launcher: Optionally, an instance implementing the `MasterLauncher` interface can be given, which
                             implements the strategy for launching the master instances in the cluster. If None, then
                             the default of the concrete cluster implementation will be chosen.
            worker_launcher: Optionally, an instance implementing the `WorkerLauncher` interface can be given, which
                             implements the strategy for launching the worker instances. If None, then the default of
                             the concrete cluster implementation will be chosen.
        """
        # Create the Logger
        self.log = logging.getLogger(__name__)

        self._group = runtime_group
        self._ports = ports if ports else list(range(self.DEFAULT_PORT_RANGE_START, self.DEFAULT_PORT_RANGE_END))
        self._master_launcher = master_launcher
        self._worker_launcher = worker_launcher

        # Cleanup will be done atexit since usage of destructor may lead to exceptions
        atexit.register(self.cleanup)

        self.log.debug('MasterWorkerCluster initialized.')

    def __str__(self):
        return type(self).__name__ + ' with ' + str(self._group)

    @property
    def master_port(self) -> int:
        """The port where the master instance was started. None, if not yet started.

        Returns:
            int: The master port.
        """
        return self._master_launcher.port

    @property
    def runtime_group(self) -> RuntimeGroup:
        """The RuntimeGroup.

        Returns:
            RuntimeGroup: The used group.
        """
        return self._group

    def start(self, worker_count: Optional[int] = None, master_port: Optional[int] = None, debug: bool = False):
        """Convenient method for launching the cluster.

        Internally, `self.start_master()` and `self.start_workers()` will be called.

        Args:
            master_port: Port of the cluster master. Will be passed on to `self.start()`, hence see respective method
                         for further details.
            worker_count: The number of worker instances to be started in the cluster. Will be passed on to
                          `self.start()`, hence see respective method for further details.
            debug: If `True`, stdout/stderr from the runtime will be printed to stdout of localhost. If, `False` then
                   the stdout/stderr will be added to python logger with level debug after each `RuntimeTask` step. Defaults to
                   `False`.

        """
        self.log.info('Starting the cluster ...')
        self.start_master(master_port, debug=debug)
        self.start_workers(worker_count, debug=debug)

    def start_master(self, master_port: Optional[int] = None, timeout: int = 3, debug: bool = False):
        """Start the master instance.

        Note:
            How the master is actually started is determined by the the actual `MasterLauncher` implementation. Another
            implementation adhering to the `MasterLauncher` interface can be provided in the constructor of the cluster
            class.

        Args:
            master_port: Port of the master instance. Defaults to self.DEFAULT_MASTER_PORT, but another one is chosen if
                         the port is not free within the group. The actual chosen port can be requested via
                         self.master_port.
            timeout: Timeout (s) after which an MasterStartError is raised if master instance not started yet.
            debug: If `True`, stdout/stderr from the runtime will be printed to stdout of localhost. Has no effect for
                   if the master instance is started locally, what default MasterLauncher implementations usually do.

        Raises:
            PortInUseError: If a single port is given and it is not free in the `RuntimeGroup`.
            NoPortsLeftError: If there are no free ports left in the port list for instantiating the master.
            MasterStartError: If master was not started after the specified `timeout`.
        """
        # 1. Determine a port or port list
        overwrite_port_list = False
        if master_port:
            ports = master_port
        elif self._group.has_free_port(self.DEFAULT_MASTER_PORT):
            ports = self.DEFAULT_MASTER_PORT
        else:
            ports = self._ports
            overwrite_port_list = True

        # 2. Trigger the actual logic for starting the master instance
        ports = self._master_launcher.start(ports, timeout, debug)  # Raises the possible exceptions

        if overwrite_port_list:
            self._ports = ports

        # Some attributes must be set in the given MasterLauncher implementation after
        # starting the master to ensure correct behavior of MasterWorkerCluster
        # => indicates a wrong implementation of the given launcher class
        assert self._master_launcher.port
        if not self._master_launcher.process:
            self.log.debug('No self._master_launcher.process is set after starting the cluster master. If the master'
                           'instance was not started as a deamon, then this could indicate buggy implementation. The'
                           'variable should be set to be able to eventually shutdown the cluster.')

        self.log.info(f'Master instance started on port {str(self.master_port)}.')

    def start_workers(self, count: Optional[int] = None, debug: bool = False):
        """Start the worker instances.

        Note:
            How workers are actually started is determined by the the actual `WorkerLauncher` implementation. Another
            implementation adhering to the `WorkerLauncher` interface can be provided in the constructor of the cluster
            class.

        Args:
            count: The number of worker instances to be started in the cluster. Defaults to the number of runtimes in
                   the cluster.
            debug: If `True`, stdout/stderr from the runtime will be printed to stdout of localhost. If, `False` then
                   the stdout/stderr will be added to python logger with level debug after each `RuntimeTask` step. Defaults to
                   `False`.
         Raises:
            NoPortsLeftError: If there are no free ports left in the port list for instantiating new worker entities.
        """
        if not count:
            count = self._group.runtime_count

        self._ports = self._worker_launcher.start(count, self.master_port, self._ports, debug)

        # Some attributes must be set in the given MasterLauncher implementation after
        # starting the master to ensure correct behavior of MasterWorkerCluster
        # => indicates a wrong implementation of the given launcher class
        # e.g. for DASK -> assert self._worker_launcher.ports_per_host

        self.log.info('Worker instances started.')

    def print_log(self):
        """Print the execution log.

        Note:
            This method is a convenient wrapper for the equivalent method of the contained `RuntimeGroup`.

        """
        self.runtime_group.print_log()

    def cleanup(self):
        """Release all resources.
        """
        self.log.info('Shutting down the MasterWorkerCluster ...')
        self._worker_launcher.cleanup()
        self._master_launcher.cleanup()
        self._group.cleanup()


import os
import time
import logging


class ExecutionFileLogUtil(object):
    """Generic class used to write log files.
    """

    def __init__(self, runtime_host, taskname):
        """Initialization method.

        Note:
            The log file file be placed in directory named `runtime_host` within the `Environment.main_directory`.

        Args:
            runtime_host: The host of the `Runtime`, where the execution takes place.
            taskname: The name of the `RuntimeTask` to be executed.
        """
        self.runtime_host = runtime_host
        self.taskname = taskname
        self.file_extension = 'log'
        self.log = logging.getLogger(__name__)
        self._main_dir = os.path.join(Environment.main_directory, 'execution_log')
        self._creation_timestamp = Timestamp()

    @property
    def file_path(self) -> str:
        """Get the full path to the log file.

        Note:
            Although, you can access the path, it does not necessary mean that it already exists. The file eventually
            gets written when the execution of the `RuntimeTask` is started.

        """
        return os.path.join(self.directory_path, f'{self.taskname}_{self._creation_timestamp.get_unformatted()}'
                                                 f'.{self.file_extension}')

    @property
    def directory_path(self) -> str:
        """Get the full path to the directory where this logfile gets written to.
        """
        path = os.path.join(self._main_dir, f'{self.runtime_host}')

        if not os.path.exists(path):
            os.makedirs(path)

        return path

    def get_write_mode(self) -> str:
        # Append if file exists otherwise create the file
        return 'a' if os.path.exists(self.file_path) else 'w+'


class Environment(object):
    """This class contains environment variables.
    """

    main_directory = os.path.abspath('./lazycluster')

    third_party_log_level = logging.CRITICAL

    use_dev_version = False

    @classmethod
    def set_main_directory(cls, dir: str):
        """Setter for the library's main directory on the manager.

        Note:
            A relative path ist also accepted and translated to an absolute path.

        Args:
            dir: Relative or absolute path.
        """
        cls.main_directory = os.path.abspath(dir)

    @classmethod
    def set_third_party_log_level(cls, log_level: int):
        """Setter for `third_party_log_level` to control the standard python logging behavior of used libraries.

        Affected libraries: paramiko

        Note:
            The class variable `third_party_log_level` defaults to `logging.Error`, e.g. only paramiko errors
            will be shown.

        Args:
            log_level: Standard python log level values as defined in `logging` like `logging.ERROR`.
        """
        logging.getLogger('paramiko').setLevel(log_level)
        logging.getLogger('invoke').setLevel(log_level)

        if log_level == logging.ERROR or log_level == logging.CRITICAL:
            logging.basicConfig(format='[%(levelname)s] %(message)s', level=logging.INFO)
        else:
            logging.basicConfig(format='[%(levelname)s] %(name)s %(message)s', level=logging.INFO)


    @classmethod
    def use_lazycluster_dev_version(cls):
        """This methods makes sure that the latest lazycluster developement version will be installed on the Runtimes. 
        This means the latest commit in the develop branch will be installed on the Runtimes. 

        Note:
            Please make sure that you install the same version on the manager as well. 
        """
        cls.use_dev_version = True


class Timestamp(object):
    """Custom Timestamp class with convenient methods.
    """

    def __init__(self):
        """ Initializes the object with the current date/time.
        """
        self._seconds_since_epoch = time.time()

        # Convert seconds since epoch to struct_time
        self._time_obj = time.localtime(self._seconds_since_epoch)

        # Ensure that each field has a fixed number of letters, so needed for representations w/o delimiters
        self.year = str(self._time_obj.tm_year)
        self.month = str(self._time_obj.tm_mon) if len(str(self._time_obj.tm_mon)) == 2 else f'0{str(self._time_obj.tm_mon)}'
        self.day = str(self._time_obj.tm_mday) if len(str(self._time_obj.tm_mday)) == 2 else f'0{str(self._time_obj.tm_mday)}'

        self.hour = str(self._time_obj.tm_hour) if len(str(self._time_obj.tm_hour)) == 2 else f'0{str(self._time_obj.tm_hour)}'
        self.min = str(self._time_obj.tm_min) if len(str(self._time_obj.tm_min)) == 2 else f'0{str(self._time_obj.tm_min)}'
        self.sec = str(self._time_obj.tm_sec) if len(str(self._time_obj.tm_sec)) == 2 else f'0{str(self._time_obj.tm_sec)}'

    def get_unformatted(self) -> str:
        """Fixed length representation w/o delimiters in format: yyyymmddhhmmss.
        """
        return self.year + self.month + self.day + self.hour + self.min + self.sec

    def get_formatted(self) -> str:
        """Formatted fixed length representation with delimiters in format: yyyy-mm-dd hh:mm:ss.
        """
        return f'{self.year}-{self.month}-{self.day} {self.hour}:{self.min}:{self.sec}'



"""Module for conveniently managing a DASK cluster. http://distributed.dask.org
"""

from distributed import Client
import time
from typing import List, Union, Optional

from subprocess import Popen

from lazycluster import RuntimeTask, Runtime, RuntimeGroup
from lazycluster.cluster import MasterWorkerCluster, MasterLauncher, WorkerLauncher
from lazycluster import _utils
from lazycluster.cluster.exceptions import MasterStartError
from lazycluster.exceptions import PortInUseError


class LocalMasterLauncher(MasterLauncher):
    """Concrete implementation of the `MasterLauncher` interface. See its documentation to get a list of the
    inherited methods and attributes.

    This class implements the logic for starting a the DASK master instance (i.e. scheduler in DASK terms) on localhost.
    """

    def start(self, ports: Union[List[int], int], timeout: int = 3, debug: bool = False) -> List[int]:
        """Launch a master instance.

        Note:
            If you create a custom subclass of MasterLauncher which will not start the master instance on localhost
            then you should pass the debug flag on to `execute_task()` of the `RuntimeGroup` or `Runtime` so that you
            can benefit from the debug feature of `RuntimeTask.execute()`.

        Args:
            ports: Port where the master should be started. If a list is given then the first port that is free in the
                   `RuntimeGroup` will be used. The actual chosen port can be requested via the property `port`.
            timeout: Timeout (s) after which an MasterStartError is raised if master instance not started yet.
            debug: If `True`, stdout/stderr from the runtime will be printed to stdout of localhost. If, `False` then
                   the stdout/stderr will be added to python logger with level debug after each `RuntimeTask` step. Defaults to
                   `False`.

        Returns:
            List[int]: In case a port list was given the updated port list will be returned. Otherwise an empty list.

        Raises:
            PortInUseError: If a single port is given and it is not free in the `RuntimeGroup`.
            NoPortsLeftError: If a port list was given and none of the ports is actually free in the `RuntimeGroup`.
            MasterStartError: If master was not started after the specified `timeout`.
        """

        if debug:
            self.log.debug('The debug flag has no effect in LocalMongoLauncher.')

        if not isinstance(ports, list):
            if _utils.localhost_has_free_port(ports) and \
               self._group.has_free_port(ports, exclude_hosts=Runtime.LOCALHOST):
                master_port = ports
            else:
                raise PortInUseError(ports, self._group)

        else:
            master_port = self._group.get_free_port(ports)  # Raises NoPortsLeftError
            ports = _utils.get_remaining_ports(ports, master_port)

        self._process = Popen(['dask-scheduler', '--port', str(master_port)])

        time.sleep(timeout)  # Needed for being able to check the port

        if not _utils.localhost_has_free_port(master_port):
            self._port = master_port
            self.log.info('Dask scheduler started on localhost on port ' + str(self._port))
        else:
            self.log.debug('Dask scheduler could NOT be started successfully on port ' + str(self._port))
            cause = f'The master port {master_port} is still free when checking after the timeout of {timeout} seconds.'
            raise MasterStartError('localhost', master_port, cause)

        # Sets up ssh tunnel for scheduler such that all communication is routed over the
        # local machine and all entities can talk to each the scheduler on localhost.
        self.log.debug(f'Expose the Dask scheduler port in the RuntimeGroup.')
        self._group.expose_port_to_runtimes(self._port)

        return ports

    def cleanup(self):
        """Release all resources.
        """
        self.log.info('Cleanup the LocalMasterLauncher ...')
        super().cleanup()


class RoundRobinLauncher(WorkerLauncher):
    """WorkerLauncher implementation for launching DASK workers in a round robin manner. See its documentation to get
    a list of the inherited methods and attributes.
    """

    def __init__(self, runtime_group: RuntimeGroup):
        """Initialization method.

        Args:
            runtime_group: The group where the workers will be started.
        """
        super().__init__(runtime_group)
        self._ports = None

        self.log.debug('RoundRobinLauncher initialized.')

    def start(self, worker_count: int, master_port: int, ports: List[int], debug: bool = False) -> List[int]:
        """Launches the worker instances in the `RuntimeGroup`.

        Args:
            worker_count: The number of worker instances to be started in the group.
            master_port:  The port of the master instance.
            ports: The ports to be used for starting the workers. Only ports from the list will be chosen
                               that are actually free.
            debug: If `True`, stdout/stderr from the runtime will be printed to stdout of localhost. If, `False` then
                   the stdout/stderr will be added to python logger with level debug after each `RuntimeTask` step. Defaults to
                   `False`.

        Returns:
            List[int]: The updated port list after starting the workers, i.e. the used ones were removed.

        Raises:
            NoPortsLeftError: If there are not enough free ports for starting all workers.
        """
        self._ports = ports
        hosts = self._group.hosts
        runtimes = self._group.runtimes

        # Launch each desired worker one by one
        for worker_index in range(worker_count):
            # Determine the runtime where the next worker will be started in
            runtime_index = (self._group.runtime_count + worker_index) % self._group.runtime_count
            # Get the actual host corresponding to the index
            host = hosts[runtime_index]
            working_dir = runtimes[runtime_index].working_dir
            assert host == runtimes[runtime_index].host

            self.log.debug(f'Launch Dask worker with index {worker_index} on Runtime {host}')

            worker_port = self._launch_single_worker(host, worker_index, master_port, working_dir, debug)  # NoPortsLeftError
            # Remember which worker ports are now used per `Runtime`
            if host in self._ports_per_host:
                self._ports_per_host[host].append(worker_port)
            else:
                self._ports_per_host.update({host: [worker_port]})

        # Needed for worker communication between workers, which is necessary in case of DASK
        self.setup_worker_ssh_tunnels()

        return self._ports

    def _launch_single_worker(self, host: str, worker_index: int, master_port: int, working_dir: str, debug: bool):
        """Launch a single worker instance in a `Runtime` in the `RuntimeGroup`.

        Raises:
            NoPortsLeftError
        """
        # 1. Get a free port based on the port list
        worker_port = self._group.get_free_port(self._ports)  # Raises NoPortsLeftError
        self._ports = _utils.get_remaining_ports(self._ports, worker_port)
        # 2. Start the worker on this port
        task = RuntimeTask('launch-dask-worker-' + str(worker_index))
        task.run_command(DaskCluster.PIP_INSTALL_COMMAND)
        task.run_command(self._get_launch_command(master_port, worker_port, working_dir))
        self._group.execute_task(task, host, omit_on_join=True, debug=debug)
        return worker_port

    @classmethod
    def _get_launch_command(cls, master_port: int, worker_port: int, working_directory: str) -> str:
        """Get the shell command for starting a worker instance.

        Returns:
             str: The launch command.
        """
        return 'dask-worker --worker-port=' + str(worker_port) + ' --local-directory=' + working_directory \
               + ' localhost:' + str(master_port)

    def cleanup(self):
        """Release all resources.
        """
        self.log.info('Cleanup the RoundRobinLauncher ...')
        super().cleanup()


class DaskCluster(MasterWorkerCluster):
    """Convenient class for launching a Dask cluster in a `RuntimeGroup`.

    DaskCluster inherits from MasterWorkerCluster. See its documentation to get a list of the inherited methods
    and attributes.

    The number of DASK workers defaults to the number of `Runtimes` in the used `RuntimeGroup`. This number can be
    adjusted so that more or less workers than available `Runtimes` can be used. Per default the desired number of
    workers is started in a round robin way as implemented in `RoundRobinLauncher`. Consequently, this leads to an
    equal distribution of DASK workers in the `RuntimeGroup`. You can provide a custom implementation inheriting from
    the `WorkerLauncher` class in order to execute a different strategy how workers should be started. The
    DASK master (i.e. scheduler) will always be started on localhost as implemented in `LocalMasterLauncher`. This
    behavior can also be changed by providing a custom implementation inheriting from the `MasterLauncher`.
    """

    DEFAULT_MASTER_PORT = 8786
    PIP_INSTALL_COMMAND = 'pip install -q "dask[complete]"'

    def __init__(self, runtime_group: RuntimeGroup, ports: Optional[List[int]] = None,
                 master_launcher: Optional[MasterLauncher] = None,
                 worker_launcher: Optional[WorkerLauncher] = None):
        """Initialization method.

        Args:
            runtime_group: The `RuntimeGroup` contains all `Runtimes` which can be used for starting the DASK entities.
            ports: The list of ports which will be used to instantiate a cluster. Defaults to
                                        list(range(self.DEFAULT_PORT_RANGE_START,
                                                   self.DEFAULT_PORT_RANGE_END)).
            master_launcher: Optionally, an instance implementing the `MasterLauncher` interface can be given, which
                             implements the strategy for launching the master instances in the cluster. If None, then
                             `LocalMasterLauncher` is used.
            worker_launcher: Optionally, an instance implementing the `WorkerLauncher` interface can be given, which
                             implements the strategy for launching the worker instances. If None, then
                             `RoundRobinLauncher` is used.
        """
        super().__init__(runtime_group, ports)

        self._master_launcher = master_launcher if master_launcher else LocalMasterLauncher(runtime_group)
        self._worker_launcher = worker_launcher if worker_launcher else RoundRobinLauncher(runtime_group)

        self.log.debug('DaskCluster initialized.')

    def get_client(self, timeout: int = 2) -> Client:
        """Get a connected Dask client. 
        
        Args:
            timeout: The timeout (s) value passed on to the Dask `Client` constructor. Defaults to 2.

        Raises:
            TimeoutError: If client connection `timeout` expires.
        """
        return Client('localhost:' + str(self.master_port), timeout=timeout)  # Raises TimeoutError

    def cleanup(self):
        """Release all resources.
        """
        self.log.info('Cleanup the RoundRobinLauncher ...')
        super().cleanup()


"""Exception module for cluster classes.

Note:
    It is intended that the lazycluster.exception.LazyclusterError should be the parent class of all defined exception
    classes here.
"""

from lazycluster.exceptions import LazyclusterError


class MasterStartError(LazyclusterError):
    """Error indicating that the cluster master instance could not be started successfully. """

    def __init__(self, host: str, port: int, cause: str):
        """Initialization method.

        Args:
            host: The host where the cluster master instance should be started.
            port: The port of the cluster master instance.
            cause: More detailed information of the actual root
        """
        self.host = host
        self.port = port
        self.cause = cause
        super().__init__(f'The cluster master instance could not be started successfully on host `{host}` on port '
                         f'{str(port)}. Cause: {cause}')



"""
Parses source code to generate API docs in markdown.
"""

import inspect
import os
import re
import sys
from inspect import getdoc, getsourcefile, getsourcelines, getmembers
from importlib import reload

if sys.version[0] == '2':
    reload(sys)
    sys.setdefaultencoding('utf8')

_RE_BLOCKSTART = re.compile(
    r"(Args:|Arg:|Arguments:|Parameters:|Kwargs:|Attributes:|Returns:|Yields:|Kwargs:|Raises:|Notes:|Note:|Examples:|Example:)",
    re.IGNORECASE)

_RE_ARGSTART = re.compile(r"(\w*?)\s*?\((.*?)\):(.*)", re.IGNORECASE)
_RE_EXCSTART = re.compile(r"(\w*?):(.*)", re.IGNORECASE)

#
# String templates
#

FUNC_TEMPLATE = \
"""-------------------
<span style="float:right;">[[source]]({path})</span>

{section} {header}

```python
{funcdef}
```
{doc}
"""

CLASS_TEMPLATE = \
"""-------------------
<span style="float:right;">[[source]]({path})</span>

{section} {header} class
{doc}
{variables}
{init}
{handlers}
{methods}
"""

MODULE_TEMPLATE = \
"""
**Source:** {path}
{global_vars}
{functions}
{classes}
"""


def get_function_signature(function, owner_class=None, show_module=False):
    isclass = inspect.isclass(function)

    # Get base name.
    name_parts = []
    if show_module:
        name_parts.append(function.__module__)
    if owner_class:
        name_parts.append(owner_class.__name__)
    if hasattr(function, '__name__'):
        name_parts.append(function.__name__)
    else:
        name_parts.append(type(function).__name__)
        name_parts.append('__call__')
        function = function.__call__
    name = '.'.join(name_parts)

    if isclass:
        function = getattr(function, '__init__', None)
    if hasattr(inspect, 'signature'):
        sig = str(inspect.signature(function))
    else:
        try:
            argspec = inspect.getargspec(function)
        except TypeError:
            # handle Py2 classes that don't define __init__
            args = ['self']
        else:
            # Generate the argument list that is separated by colons.
            args = argspec.args[:]
            if argspec.defaults:
                offset = len(args) - len(argspec.defaults)
                for i, default in enumerate(argspec.defaults):
                    args[i + offset] = '{}={!r}'.format(args[i + offset], argspec.defaults[i])
            if argspec.varargs:
                args.append('*' + argspec.varargs)
            if argspec.keywords:
                args.append('**' + argspec.keywords)
        sig = '(' + ', '.join(args) + ')'

    return name + sig


def make_iter(obj):
    """ Makes an iterable
    """
    return obj if hasattr(obj, '__iter__') else [obj]


def order_by_line_nos(objs, line_nos):
    """Orders the set of `objs` by `line_nos`
    """
    ordering = sorted(range(len(line_nos)), key=line_nos.__getitem__)
    return [objs[i] for i in ordering]


def to_md_file(string, filename, out_path="."):
    """Import a module path and create an api doc from it
    Args:
        string (str): string with line breaks to write to file.
        filename (str): filename without the .md
        out_path (str): The output directory
    """
    md_file = "%s.md" % filename
    with open(os.path.join(out_path, md_file), "w") as f:
        f.write(string)
    print("wrote {}.".format(md_file))


def modules2mdfiles():
    # TODO
    pass


def code_snippet(snippet):
    result = '```python\n'
    result += snippet + '\n'
    result += '```\n'
    return result


class MarkdownAPIGenerator(object):
    def __init__(self, src_root, github_link):
        """Initializes the markdown api generator.
        Args:
            src_root: The root folder name containing all the sources.
                Ex: src
            github_link: The base github link. Should include branch name.
                Ex: https://github.com/raghakot/keras-vis/tree/master
                All source links are generated with this prefix.
        """
        self.src_root = src_root
        self.github_link = github_link

    def get_line_no(self, obj):
        """Gets the source line number of this object. None if `obj` code cannot be found.
        """
        try:
            lineno = getsourcelines(obj)[1]
        except:
            # no code found
            lineno = None
        return lineno

    def get_src_path(self, obj, append_base=True):
        """Creates a src path string with line info for use as markdown link.
        """
        path = getsourcefile(obj)
        if not self.src_root in path:
            # this can happen with e.g.
            # inlinefunc-wrapped functions
            if hasattr(obj, "__module__"):
                path = "%s.%s" % (obj.__module__, obj.__name__)
            else:
                path = obj.__name__
            path = path.replace(".", "/")
        print(path)
        pre, post = path.rsplit(self.src_root + "/", 1)

        lineno = self.get_line_no(obj)
        lineno = "" if lineno is None else "#L{}".format(lineno)

        path = self.src_root + "/" + post + lineno
        if append_base:
            path = os.path.join(self.github_link, path)
        return path

    def doc2md(self, func):
        """Parse docstring (parsed with getdoc) according to Google-style
        formatting and convert to markdown. We support the following
        Google style syntax:
        Args, Kwargs:
            argname (type): text
            freeform text
        Returns, Yields:
            retname (type): text
            freeform text
        Raises:
            exceptiontype: text
            freeform text
        Notes, Examples:
            freeform text
        """
        doc = getdoc(func) or ""
        blockindent = 0
        argindent = 1
        out = []

        for line in doc.split("\n"):
            indent = len(line) - len(line.lstrip())
            line = line.lstrip()
            if _RE_BLOCKSTART.match(line):
                # start of a new block
                blockindent = indent
                out.append("\n**{}**\n".format(line))
            elif indent > blockindent:
                if _RE_ARGSTART.match(line):
                    # start of new argument
                    out.append("\n" + " " * blockindent + " - " + _RE_ARGSTART.sub(r"`\1` (\2): \3", line))
                    argindent = indent
                elif _RE_EXCSTART.match(line):
                    # start of an exception-type block
                    out.append("\n" + " " * blockindent + " - " + _RE_EXCSTART.sub(r"`\1`: \2", line))
                    argindent = indent
                elif indent > argindent:
                    out.append("\n" + " " * (blockindent + 2) + line)
                else:
                    out.append("\n" + line)
            else:
                out.append("\n" + line)

            out.append("\n")

        return "".join(out)

    def func2md(self, func, clsname="", depth=3):
        """Takes a function (or method) and documents it.
        Args:
            clsname (str, optional): class name to prepend to funcname.
            depth (int, optional): number of ### to append to function name
        """
        section = "#" * depth
        funcname = func.__name__
        escfuncname = "`%s`" % funcname if funcname.startswith("_") else funcname
        header = "%s%s" % ("%s." % clsname if clsname else "", escfuncname)

        path = self.get_src_path(func)
        doc = self.doc2md(func)

        funcdef = get_function_signature(func)
        # -> is rendered with a red background in github readme
        funcdef = funcdef.replace("->", "→")
        funcdef = funcdef.replace("=", " = ")
        funcdef = funcdef.replace(":", ": ")

        # split the function definition if it is too long
        lmax = 70
        if len(funcdef) > lmax:
            import re
            param_part = re.findall("\((.*?)\)", funcdef)
            if len(param_part) > 0:
                param_part = param_part[0]
            param_str = ""
            params = param_part.split(", ")
            for i, param in enumerate(params):
                if i is not len(params) - 1:
                    param_str += "    " + str(param) + ",\n"
                else:
                    param_str += "    " + str(param) + "\n"

            # wrap in the args list
            first_part = funcdef.split("(", 1)[0]
            last_part = funcdef.split(")", 1)[1]

            funcdef = first_part + "(\n"
            funcdef += param_str
            funcdef += ")" + last_part

        # build the signature
        string = FUNC_TEMPLATE.format(section=section,
                                      header=header,
                                      funcdef=funcdef,
                                      path=path,
                                      doc=doc if doc else "*No documentation found.*")
        return string

    def class2md(self, cls, depth=2):
        """Takes a class and creates markdown text to document its methods and variables.
        """

        section = "#" * depth
        subsection = "#" * (depth + 2)
        clsname = cls.__name__
        modname = cls.__module__
        header = clsname
        path = self.get_src_path(cls)
        doc = self.doc2md(cls)

        try:
            init = self.func2md(cls.__init__, clsname=clsname)
        except (ValueError, TypeError):
            # this happens if __init__ is outside the repo
            init = ""

        variables = []
        for name, obj in getmembers(cls, lambda a: not (inspect.isroutine(a) or inspect.ismethod(a))):
            if not name.startswith("_") and type(obj) == property:
                comments = self.doc2md(obj) or inspect.getcomments(obj)
                comments = "\n %s" % comments if comments else ""
                variables.append("\n%s %s.%s%s\n" % (subsection, clsname, name, comments))

        handlers = []
        for name, obj in getmembers(cls, inspect.ismethoddescriptor):
            if not name.startswith("_") and hasattr(obj, "__module__"): # and obj.__module__ == modname:
                handlers.append("\n%s %s.%s\n *Handler*" % (subsection, clsname, name))

        methods = []
        for name, obj in getmembers(cls, inspect.isfunction):
            print(name)
            if not name.startswith("_") and hasattr(obj,
                                                    "__module__") and name not in handlers: #and obj.__module__ == modname and :
                methods.append(self.func2md(obj, clsname=clsname, depth=depth + 1))

        string = CLASS_TEMPLATE.format(section=section,
                                       header=header,
                                       path=path,
                                       doc=doc if doc else "",
                                       init=init,
                                       variables="".join(variables),
                                       handlers="".join(handlers),
                                       methods="".join(methods))
        return string

    def module2md(self, module):
        """Takes an imported module object and create a Markdown string containing functions and classes.
        """
        modname = module.__name__
        path = self.get_src_path(module, append_base=False)
        path = "[{}]({})".format(path, os.path.join(self.github_link, path))
        found = []

        classes = []
        line_nos = []
        for name, obj in getmembers(module, inspect.isclass):
            # handle classes
            found.append(name)
            if not name.startswith("_") and hasattr(obj, "__module__") and obj.__module__ == modname:
                classes.append(self.class2md(obj))
                line_nos.append(self.get_line_no(obj) or 0)
        classes = order_by_line_nos(classes, line_nos)

        functions = []
        line_nos = []
        for name, obj in getmembers(module, inspect.isfunction):
            # handle functions
            found.append(name)
            if not name.startswith("_") and hasattr(obj, "__module__") and obj.__module__ == modname:
                functions.append(self.func2md(obj))
                line_nos.append(self.get_line_no(obj) or 0)
        functions = order_by_line_nos(functions, line_nos)

        variables = []
        line_nos = []
        for name, obj in module.__dict__.items():
            if not name.startswith("_") and name not in found:
                if hasattr(obj, "__module__") and obj.__module__ != modname:
                    continue
                if hasattr(obj, "__name__") and not obj.__name__.startswith(modname):
                    continue
                comments = inspect.getcomments(obj)
                comments = ": %s" % comments if comments else ""
                variables.append("- **%s**%s" % (name, comments))
                line_nos.append(self.get_line_no(obj) or 0)

        variables = order_by_line_nos(variables, line_nos)
        if variables:
            new_list = ["**Global Variables**", "---------------"]
            new_list.extend(variables)
            variables = new_list

        string = MODULE_TEMPLATE.format(path=path,
                                        global_vars="\n".join(variables) if variables else "",
                                        functions="\n".join(functions) if functions else "",
                                        classes="".join(classes) if classes else "")
        return string


"""Runtime management module. This module contains convenient classes for working with `Runtimes` and `RuntimeTasks`."""

from typing import Optional, List, Dict, Generator, Union
from storm import Storm
from multiprocessing import Process, Pool
import warnings
import logging
import atexit
from copy import deepcopy

from lazycluster import RuntimeTask, Runtime
from lazycluster import _utils
from lazycluster import PortInUseError, InvalidRuntimeError, NoPortsLeftError, NoRuntimesDetectedError


class RuntimeGroup(object):
    """A `RuntimeGroup` is the representation of logically related `Runtimes` and provides convenient methods for
    managing those related `Runtimes`. Most methods are wrappers around their counterparts in the `Runtime` class.
    Typical usage examples are exposing a port (i.e. a service such as a DB) in the `RuntimeGroup`, transfer files, or
    execute  a `RuntimeTask` on the contained `Runtimes`. Additionally, all concrete RuntimeCluster
    (e.g. HyperoptCluster) implementations rely on `RuntimeGroups` for example.

    Examples:
            Execute a `RuntimeTask` in a `RuntimeGroup`
            ```python
            # Create instances
            group = RuntimeGroup([Runtime('host-1'), Runtime('host-2')])
            # group = RuntimeGroup(hosts=['host-1', 'host-2'])
            my_task = RuntimeTask('group-demo').run_command('echo Hello Group!')

            # Execute a RuntimeTask in a single Runtime
            single_task = group.execute_task(my_task)
            print(single_task.execution_log[0])

            # Execute a RuntimeTask in the whole RuntimeGroup
            task_list = group.execute_task(my_task, broadcast=True)

            # Execute a RuntimeTask on a single Runtime contained in the RuntimeGroup
            task = group.execute_task(my_task)
            ```
            A DB is running on localhost on port `local_port` and the DB is only accessible
            from localhost. But you also want to access the service on the other `Runtimes` on port
            `runtime_port`. Then you can use this method to expose the service which is running on the
            local machine to the remote machines.
            ```python
            # Expose a port to all Runtimes contained in the Runtime. If a port list is given the next free port is
            # chosen and returned.
            group_port = group.expose_port_to_runtimes(local_port=60000, runtime_port=list(range(60000,60010)))
            print('Local port 60000 is now exposed to port ' + str(group_port) + ' in the RuntimeGroup!')
            ```
            A DB is running on a remote host on port `runtime_port` and the DB is only accessible from the remote
            machine itself. But you also want to access the service to other `Runtimes` in the group. Then you can use
            this method to expose the service which is running on one `Runtime` to the whole group.
            ```python
            # Expose a port from a Runtime to all other ones in the RuntimeGroup. If a port list is given the next
            # free port is chosen and returned.
            group_port = group.expose_port_from_runtime_to_group(host='host-1', runtime_port=60000,
                                                                 group_port=list(range(60000,60010)))
            print('Port 60000 of `host-1` is now exposed to port ' + str(group_port) + ' in the RuntimeGroup!')
            ```
    """

    _INTERNAL_PORT_MIN = 5800
    _INTERNAL_PORT_MAX = 5999
    _internal_port_range = range(_INTERNAL_PORT_MIN, _INTERNAL_PORT_MAX)

    def __init__(self, runtimes: Optional[List[Runtime]] = None, hosts: Optional[List[str]] = None,
                 working_dir: Optional[str] = None):
        """Initialization method.
        
        Args:
            runtimes: List of `Runtimes`. If not given, then `hosts` must be supplied.
            hosts: List of hosts, which will be used to instantiate `Runtime` objects. If not given, then `runtimes`
                   must be supplied.
            working_dir: The directory which shall act as working directory on all Runtimes. Defaults to None. See the `Runtime` docs for
                         further details.

        Raises:
            ValueError: Either `runtimes` or `hosts` must be supplied. Not both or none.
            InvalidRuntimeError: If a runtime cannot be instantiated via host.
        """
        if not runtimes and not hosts or runtimes and hosts:
            raise ValueError("Either `runtimes` or `hosts` must be supplied. Not both or none.")

        # Create the Logger
        self.log = logging.getLogger(__name__)

        if runtimes:
            self._runtimes = {runtime.host: runtime for runtime in runtimes}
        elif hosts:
            self._runtimes = {host: Runtime(host) for host in hosts}  # Throws InvalidRuntimeError
        self._proc_keys = []
        self._tasks = []
        # Cleanup will be done atexit since usage of destructor may lead to exceptions
        atexit.register(self.cleanup)

        self.log.debug(f'RuntimeGroup initialized.')

    def __repr__(self):
        return "%s(%r)" % (self.__class__, self.__dict__)

    def __str__(self):
        hosts_str = ''
        for runtime in self.runtimes:
            if not hosts_str:
                hosts_str += str(runtime)
            else:
                hosts_str += ', ' + str(runtime)
        return type(self).__name__ + ': ' + hosts_str

    @property
    def hosts(self) -> List[str]:
        """Contained hosts in the group.
        
        Returns:
            List[str]: List with hosts of all `Runtimes`.
        """
        return list(self._runtimes)

    @property
    def runtimes(self) -> List[Runtime]:
        """Contained Runtimes in the group.

        Returns:
            List[Runtime]: List with all `Runtimes`.
        """
        return list(self._runtimes.values())

    @property
    def runtime_count(self) -> int:
        """The count of runtimes contained in the group.

        Returns:
            int: The count.
        """
        return len(self.hosts)

    @property
    def task_processes(self) -> List[Process]:
        """Processes from all contained `Runtimes` which were started to execute a `RuntimeTask`.

        Returns:
             List[Process]: Process list.
        """
        processes = []
        for runtime in self._runtimes.values():
            processes.extend(runtime.task_processes)
        return processes

    @property
    def function_returns(self) -> Generator[object, None, None]:
        """Function return data. Blocks thread until a `RuntimeTasks` finished executing and gives back the return data
        of the remotely executed python functions. The data is returned in the same order as the Tasks were started.

        Note:
            Only function returns from `RuntimeTasks` that were started via the `RuntimeGroup` will be returned. If a
            contained `Runtime` executed further `RuntimeTasks` directly, then those data will only be returned when querying the
            respective `RuntimeTask` directly.

        Returns:
            Generator[object, None, None]: The unpickled return data.
        """
        self.log.debug(f'Start generating function returns for RuntimeGroup.')
        for task in self._tasks:
            for return_data in task.function_returns:
                yield return_data

    def set_env_variables(self, env_variables: Dict):
        """Set environment variables used when executing a task.

        Note:
            This is a convenient wrapper and internally uses Runtime.env_variables.

        Args:
            env_variables: The env variables as dictionary.
        """
        for runtime in self._runtimes.values():
            runtime.env_variables =  env_variables

    def add_env_variables(self, env_variables: Dict):
        """Update the environment variables of all contained Runtimes. If a variable already
        exists it gets updated and if not it will be added.

        Note:
            This is a convenient wrapper and internally calls Runtime.add_env_variables().

        Args:
            env_variables: The env variables used for the update.
        """
        for runtime in self._runtimes.values():
            runtime.add_env_variables(env_variables)

    def print_hosts(self):
        """Print the hosts of the group.
        """
        if not self.hosts:
            print('The group is empty!')
            return

        for hostname in self.hosts:
            print(hostname)

    def print_runtime_info(self):
        """Print information of contained `Runtimes`.
        """
        # Ensure that all runtime info are read asynchronously before printing.
        # Otherwise each runtime will read its information synchronously.
        self.fill_runtime_info_buffers_async()

        for runtime in self.runtimes:
            runtime.print_info()
            print('\n')

    def fill_runtime_info_buffers_async(self):
        """Trigger the reading of runtime information asynchronously and buffer the result.

        Note:
            The actual reading of `Runtime.info data takes place when requesting the attribute the first time.
            Consequently, the result gets buffered in the respective Runtime instance. The actual reading of the data
            takes places on the remote host and takes some seconds. This method enables you to read the information in a
            separate processes so that the execution time stays more or less the same independent of the actual amount
            of Runtimes used.

        """
        self.log.debug('Start filling the buffers Runtime.info asynchronously.')
        with Pool(self.runtime_count) as pool:
            results = []
            for runtime in self._runtimes.values():
                # We need to check the private member because otherwise the reading will be triggered synchronously
                if runtime._info:
                    # If the runtime info is already present, we do not need to re-read it
                    continue
                self.log.debug(f'Start reading runtime info asynchronously of Runtime {runtime.host}')
                results.append(pool.apply_async(_utils.read_host_info, (runtime.host, )))

            index = 1
            for result in results:
                self.log.debug(f'Waiting fo result {index} of {self.runtime_count}.')
                index += 1
                runtime_info = result.get()
                self.log.debug(f'Result for host {runtime_info["host"]} retrieved')
                self.get_runtime(runtime_info['host'])._info = runtime_info

    def add_runtime(self, host: Optional[str] = None, runtime: Optional[Runtime] = None):
        """Add a `Runtime` to the group either by host or as a `Runtime` object.
        
        Args:
            host: The host of the runtime. Defaults to None.
            runtime: The `Runtime` object to be added to the group. Defaults to None.

        Raises:
            ValueError: If the same host is already contained. Or if both host and runtime is given. We refuse
                        the temptation to guess. Also if no argument is supplied.
        """
        if host and runtime:
            raise ValueError('Only host or runtime must be supplied. We refuse the temptation to guess.')
        elif not host and not runtime:
            raise ValueError('Hostname or runtime must be supplied.')
        elif runtime and self.contains_runtime(runtime.host):
            raise ValueError('Runtime with the same host is already contained in the group.')
        elif host and self.contains_runtime(host):
            raise ValueError('Runtime with the same host is already contained in the group.')
        elif host:
            rt = Runtime(host)
        else:
            rt = runtime
        self._runtimes[rt.host] = rt
        self.log.info(rt.host + ' added as ' + rt.class_name + ' to the group.')

    def remove_runtime(self, host: str):
        """Remove a runtime from the group by host.
        
        Args:
            host: The host of the `Runtime` to be removed from the group.
        """
        if not self.contains_runtime(host):
            warnings.warn('Runtime ' + host + ' is  not contained in the group!')
            return
        else:
            del self._runtimes[host]
            self.log.info(f'Runtime {host} removed from RuntimeGroup')

    def expose_port_to_runtimes(self, local_port: int, runtime_port: Union[int, List[int], None] = None,
                                exclude_hosts: Union[str, List[str], None] = None) -> int:
        """ Expose a port from localhost to all Runtimes beside the excluded ones so that all traffic on the
        `runtime_port` is forwarded to the `local_port` on the local machine. This corresponds to remote
        port forwarding in ssh tunneling terms. Additionally, all relevant runtimes will be checked if the port is
        actually free.
        
        Args:
            local_port: The port on the local machine.
            runtime_port: The port on the runtimes where the `local_port` shall be exposed to. May raise PortInUseError
                          if a single port is given. If a list is used to automatically find a free port then a
                          NoPortsLeftError may be raised. Defaults to `local_port`.
            exclude_hosts: List with hosts where the port should not be exposed to. Defaults to None. Consequently, all
                           `Runtimes` will be considered.

        Returns:
            int: The port which was actually exposed to the `Runtimes`.

        Raises:
            PortInUseError: If `runtime_port` is already in use on at least one Runtime.
            ValueError: Only hosts or `exclude_hosts` must be provided or host is not contained in the group.
        """

        self.log.info(f'Start exposing the local port {str(local_port)} in the RuntimeGroup.')

        # 1. Determine a free runtime port
        if not runtime_port or not isinstance(runtime_port, list):
            if not runtime_port:
                runtime_port = local_port
            if not self.has_free_port(runtime_port):
                raise PortInUseError(runtime_port, self)

        elif isinstance(runtime_port, list):
            runtime_port = self.get_free_port(runtime_port)  # Raises NoPortsLeftError

        # Ensure that we work with a list and not a single value
        exclude_hosts = _utils.create_list_from_parameter_value(exclude_hosts)

        if not self.has_free_port(runtime_port, exclude_hosts=exclude_hosts + [Runtime.LOCALHOST]):
            raise PortInUseError(runtime_port, self)

        for runtime in self.get_runtimes(exclude_hosts=exclude_hosts).values():  # Raises ValueError
            process_key = runtime.expose_port_to_runtime(local_port, runtime_port)
            if process_key:
                self._proc_keys.append(process_key)

        self.log.debug(f'Port {str(runtime_port)} used as runtime_port.')
        return runtime_port

    def expose_port_from_runtime_to_group(self, host: str, runtime_port: int,
                                          group_port: Union[int, List[int], None] = None) -> int:
        """Expose a port from a `Runtime` to all other `Runtimes` in the `RuntimeGroup` so that all traffic to the
        `group_port` is forwarded to the `runtime_port` of the runtime.
        
        Args:
            host: The host of the `Runtime`.
            runtime_port: The port on the runtime.
            group_port: The port on the other runtimes where the `runtime_port` shall be exposed to. May raise
                        PortInUseError if a single port is given. If a list is used to automatically find a free port
                        then a NoPortsLeftError may be raised. Defaults to runtime_port.

        Returns:
            int: The `group_port` that was eventually used.

        Raises:
            ValueError: If host is not contained.
            PortInUseError: If `group_port` is occupied on the local machine.
            NoPortsLeftError: If `group_ports` was given and none of the ports was free.
        """
        if not self.contains_runtime(host):
            raise ValueError('Runtime ' + host + ' is not contained in the group.')

        self.log.info(f'Start exposing the port {str(runtime_port)} from Runtime {host} in the RuntimeGroup.')

        # 1. Determine a free group port
        if not group_port or not isinstance(group_port, list):
            if not group_port:
                group_port = runtime_port
            if not self.has_free_port(group_port, exclude_hosts=[host]):
                raise PortInUseError(group_port, self)

        elif isinstance(group_port, list):
            group_port = self.get_free_port(group_port)

        # 2. Determine a free port on localhost, since all traffic is tunneled over the local machine
        local_port = None
        if _utils.localhost_has_free_port(group_port):
            local_port = group_port
        else:
            # Get the first free port from the local port range
            for current_port in self._internal_port_range:
                if not _utils.localhost_has_free_port(current_port):
                    continue
                else:
                    local_port = current_port
                    self._internal_port_range = range(current_port + 1, self._INTERNAL_PORT_MAX)

        if not local_port:
            # self._internal_port_range = None <- does this makes sense? not sure!
            raise NoPortsLeftError()

        # 3. Finally expose the ports
        for runtime in self._runtimes.values():
            if runtime.host == host:
                process_key = runtime.expose_port_from_runtime(runtime_port, local_port)
            else:
                process_key = runtime.expose_port_to_runtime(local_port, group_port)
            if process_key:
                self._proc_keys.append(process_key)
        self.log.debug(f'Port {str(group_port)} used as group_port.')
        return group_port

    def execute_task(self, task: RuntimeTask, host: Optional[str] = None, broadcast: bool = False,
                     execute_async: bool = True, omit_on_join: bool = False,
                     debug: bool = False) -> RuntimeTask or List[RuntimeTask]:
        """Execute a `RuntimeTask` in the whole group or in a single `Runtime`.

        Note:
            When broadcasting a `RuntimeTask` in the group then actually deep copies of the RuntimeTask are created (by using
            its custom __deepcopy__ implementation), since each RuntimeTask holds state related to its own execution. Thus,
            multiple `RuntimeTasks` will be returned in this case.
        
        Args:
            task: The RuntimeTask to be executed.
            host: If `RuntimeTask` should be executed in one Runtime. Optionally, the host could be set in order to ensure
                  the execution in a specific Runtime. Defaults to None. Consequently, the least busy `Runtime` will be
                  chosen.
            broadcast: True, if the `RuntimeTask` will be executed on all `Runtimes`. Defaults to False.
            execute_async: True, if execution will take place async. Defaults to True.
            omit_on_join: If True, then a call to join() won't wait for the termination of the corresponding process.
                          Defaults to False. This parameter has no effect in case of synchronous execution.
            debug : If `True`, stdout/stderr from the remote host will be printed to stdout. If, `False`
                    then the stdout/stderr will be written to execution log files. Defaults to `False`.

        Returns:
            RuntimeTask or List[RuntimeTask]: Either a single `RuntimeTask` object in case the execution took place
                                              in a single `Runtime` or a list of `RuntimeTasks` if executed in all.

        Raises:
            ValueError: If `host` is given and not contained as `Runtime` in the group.
            TaskExecutionError: If an executed `RuntimeTask` step can't be executed successfully.
        """
        if not broadcast:
            self.log.debug(f'Start executing RuntimeTask {task.name} in RuntimeGroup (no broadcasting).')
        else:
            self.log.debug(f'Start broadcasting RuntimeTask {task.name} in RuntimeGroup.')

        task.omit_on_join = omit_on_join

        if broadcast:
            tasks = []
            needs_to_create_copy = False

            for runtime in self.get_runtimes().values():  # Raises ValueError

                if not needs_to_create_copy:
                    current_task = task
                else:
                    # Create a deep copy to prevent reference errors especially for the task log.
                    # Each task will contain its own log produced on its executing host.
                    current_task = deepcopy(task)

                runtime.execute_task(current_task, execute_async, debug)

                tasks.append(current_task)
                self._tasks.append(current_task)

            return tasks

        else:
            if host:
                if host not in self._runtimes:
                    raise ValueError('The host ' + host + ' is not a valid runtime.')
                self.get_runtime(host).execute_task(task, execute_async, debug)
            else:
                self._get_least_busy_runtime().execute_task(task, execute_async, debug)
            self._tasks.append(task)
            return task

    def send_file(self, local_path: str, remote_path: Optional[str] = None, execute_async: Optional[bool] = True) \
            -> List[RuntimeTask]:
        """Send either a single file or a folder from the manager to all `Runtimes` of the group.

        Note:
            This method is a convenient wrapper around the Runtime's send file functionality. See `Runtime.send_file()´
            for further details.

        Args:
            local_path: Path to file on local machine.
            remote_path: Path on the `Runtime`. Defaults to the `Runtime.working_dir`. See
                         `RuntimeTask.execute()` docs for further details.
            execute_async: Each individual sending will be done in a separate process if True. Defaults to True.

        Returns:
            List[RuntimeTask]: The `RuntimeTasks` that were internally created by the respective `Runtimes`.

        Raises:
            ValueError: If local_path is emtpy.
            TaskExecutionError: If an executed `RuntimeTask` step can't be executed successfully.
            OSError: In case of non existent paths.
        """
        async_str = ' asynchronously ' if execute_async else ' synchronously '
        self.log.debug(f'Start sending local file `{local_path}` to RuntimeGroup {async_str}. Given remote path: '
                       f'`{remote_path}`.')

        tasks = []

        for runtime in self._runtimes.values():
            executed_task = runtime.send_file(local_path, remote_path, execute_async)
            self._tasks.append(executed_task)
            tasks.append(executed_task)

        return tasks

    def join(self):
        """Blocks until `RuntimeTasks` which were started via the `group.execute_task()` method terminated.
        """
        self.log.info('Joining all processes executing a RuntimeTask that were started via the RuntimeGroup')
        for task in self._tasks:
            task.join()

    def print_log(self):
        """Print the execution logs of the contained `Runtimes` that were executed in the group.
        """
        for runtime in self._runtimes.values():
            print(f'Execution Log of Runtime {runtime.host}:')
            runtime.print_log()

    def get_free_port(self, ports: List[int], enforce_check_on_localhost: bool = False) -> int:
        """Return the first port from the list which is currently not in use in the whole group.

        Args:
             ports: The list of ports that will be used to find a free port in the group.
             enforce_check_on_localhost: If true the port check will be executed on localhost as well, although
                                         localhost might not be a `Runtime` instance contained in the `RuntimeGroup`.

        Returns:
            int: The first port from the list which is not yet used within the whole group.

        Raises:
            NoPortsLeftError: If the port list is empty and no free port was found yet.
        """
        localhost_not_in_group = True if Runtime.LOCALHOST not in self._runtimes else False

        for port in ports:

            if enforce_check_on_localhost and localhost_not_in_group and not _utils.localhost_has_free_port(port):
                continue

            if not self.has_free_port(port):
                continue

            self.log.debug(f'Port {str(port)} is retrieved as free port in the RuntimeGroup.')
            return port

        raise NoPortsLeftError()

    def has_free_port(self, port: int, exclude_hosts: Union[List[str], str, None] = None) -> bool:
        """Check if a given port is free on `Runtimes` contained in the group. The check can be restricted to a
        specific subset of contained `Runtimes` by excluding some hosts.

        Args:
            port: The port to be checked in the group.
            exclude_hosts: If supplied, the check will be omitted in these `Runtimes`. Defaults to None, i.e. not
                           restricted.

        Returns:
            bool: True if port is free on all `Runtimes`, else False.

        Raises:
            ValueError: Only hosts or exclude_hosts must be provided or Hostname is
                        not contained in the group.                     
        """
        is_free = True

        for runtime in self.get_runtimes(exclude_hosts=exclude_hosts).values():  # Raises ValueError
            if not runtime.has_free_port(port):
                self.log.debug(f'The port {str(port)} is currently in use on {runtime.class_name} {runtime.host}')
                is_free = False

        return is_free

    def contains_runtime(self, host: str) -> bool:
        """Check if the group contains a `Runtime` identified by host.

        Args:
            host: The `Runtime` to be looked for.

        Returns:
            bool: True if runtime is contained in the group, else False.
        """
        return True if host in self.hosts else False

    def clear_tasks(self):
        """Clears all internal state related to `RuntimeTasks`.
        """
        self.log.info(f'Clear all RuntimeTasks and kill related processes in the RuntimeGroup.')
        # 1: Clean up all state on group level
        self._tasks = []
        self._proc_keys = [proc_key for proc_key in self._proc_keys if not Runtime.is_runtime_task_process(proc_key)]
        # 2: Clean up state in contained `Runtimes`
        for runtime in self._runtimes.values():
            runtime.clear_tasks()

    def get_runtime(self, host: Optional[str] = None) -> Runtime:
        """Returns a runtime identified by the `host` or 'the least busy one' will be returned if not host is given,
        i.e. the one with the fewest alive processes executing a `RuntimeTask`.

        Note:
            The current behavior of the 'least busy runtime' is intended to be changed to a smarter approach as soon as
            there will be the concrete need. So feel free to reach out to us or provide an alternative approach as PR.

        Args:
            host: The host which identifies the runtime.

        Returns:
            Runtime: Runtime object.

        Raises:
            ValueError: Hostname is not contained in the group.
        """
        if host and host not in self.hosts:
            raise ValueError('Hostname not contained in the RuntimeGroup')
        elif host:
            return self._runtimes[host]
        else:
            self._get_least_busy_runtime()

    def get_runtimes(self, include_hosts: Union[str, List[str], None] = None,
                           exclude_hosts: Union[str, List[str], None] = None) -> Dict[str, Runtime]:
        """Convenient methods for getting relevant `Runtimes` contained in the group.

        Args:
            include_hosts: If supplied, only the specified `Runtimes` will be
                                                                 returned. Defaults to None, i.e. not restricted.
            exclude_hosts: If supplied, all `Runtimes` beside the here specified ones will be returned. Defaults to an
                           empty list, i.e. not restricted.
        Raises:
            ValueError: If include_hosts and exclude_hosts is provided or if a host from `include_host` is not contained
                        in the group.
        """
        runtimes = {}

        if include_hosts and exclude_hosts:
            raise ValueError('Only include_hosts or exclude_hosts must be provided. Not both.')

        elif not include_hosts and not exclude_hosts:
            # => Return all runtimes
            return self._runtimes

        elif include_hosts:
            # => Only the ones specified in hosts
            include_hosts = _utils.create_list_from_parameter_value(include_hosts)  # Ensure it's a list
            for hostname in include_hosts:
                if hostname not in self._runtimes:
                    raise ValueError(hostname + ' is not contained in the RuntimeGroup.')
                runtimes.update({hostname: self._runtimes[hostname]})

        elif exclude_hosts:
            # => All but without the ones specified in excluded_hosts
            exclude_hosts = _utils.create_list_from_parameter_value(exclude_hosts)  # Ensure it's a list
            for hostname in self.hosts:
                if hostname not in exclude_hosts:
                    runtimes.update({hostname: self._runtimes[hostname]})

        return runtimes

    def cleanup(self):
        """Release all acquired resources and terminate all processes by calling the cleanup method on all contained
        `Runtimes`.
        """
        self.log.info(f'Start cleanup of RuntimeGroup.')
        for runtime in self.get_runtimes().values():
            runtime.cleanup()

    def _get_least_busy_runtime(self) -> Runtime:
        """Get the least busiest runtime in terms of alive `RuntimeTask` processes.
        """
        final_rt = None

        for runtime in self.get_runtimes().values():
            if not final_rt:
                final_rt = runtime
            elif runtime.alive_task_process_count < final_rt.alive_task_process_count:
                final_rt = runtime
        return final_rt


class RuntimeManager(object):
    """The `RuntimeManager` can be used for a simplified resource management, since it aims to automatically detect
    valid `Runtimes` based on the ssh configuration. It can be used to create a `RuntimeGroup` based on the
    automatically detected instances and possibly based on further filters such as GPU availability.
    """

    def __init__(self):
        """Initialization method.

        Raises:
            NoRuntimeDetectedError: If no `Runtime` could be automatically detected.
        """
        # Create the Logger
        self.log = logging.getLogger(__name__)

        runtimes = {}
        self.inactive_hosts = []

        # Iterate over ssh configuration entries and look for valid RemoteRuntimes
        ssh_util = Storm()
        self.log.debug('RuntimeManager starts looking for Runtimes based on ssh configuration.')

        for ssh_entry in ssh_util.list_entries(only_servers=True):
            if ssh_entry['host'] in runtimes:
                continue
            if (ssh_entry['host'] == 'localhost' and '127.0.0.1' in runtimes) or (ssh_entry['host'] == '127.0.0.1'
                                                                                  and 'localhost' in runtimes):
                continue
            try:
                self.log.debug(f'RuntimeManager tries to instantiate host {ssh_entry["host"]} as Runtime.')
                runtime = Runtime(ssh_entry['host'])
            except InvalidRuntimeError:
                self.inactive_hosts.append(ssh_entry['host'])
                self.log.debug(f'RuntimeManager detected host config for {ssh_entry["host"]}, that could NOT be '
                               f'instantiated  as a valid Runtime.')
                continue
            runtimes.update({runtime.host: runtime})
            self.log.info(runtime.host + ' detected as valid Runtime by the RuntimeManager.')

        try:
            self._group = RuntimeGroup(list(runtimes.values()))
            self.log.info(f'RuntimeManager detected {len(runtimes)} valid Runtime(s).')
        except ValueError as err:
            raise NoRuntimesDetectedError(err)

        self.log.debug('RuntimeManager initialized.')

    def __repr__(self):
        return "%s(%r)" % (self.__class__, self.__dict__)

    def __str__(self):
        runtimes_str = ''
        for runtime in self._group.runtimes:
            runtimes_str += ', ' + str(runtime)
        return type(self).__name__ + ' detected: ' + runtimes_str

    def create_group(self,
                     include_hosts: Union[str, List[str], None] = None,
                     exclude_hosts: Union[str, List[str], None] = None,
                     gpu_required: bool = False,
                     min_memory: Optional[int] = None,
                     min_cpu_cores: Optional[int] = None,
                     installed_executables: Union[str, List[str], None] = None,
                     filter_commands: Union[str, List[str], None] = None,
                     working_dir: Optional[str] = None) -> RuntimeGroup:
        """Create a runtime group with either all detected `Runtimes` or with a subset thereof.
        
        Args:
            include_hosts: Only these hosts will be included in the `RuntimeGroup`. Defaults to None, i.e. not
                           restricted.
            exclude_hosts: If supplied, all detected `Runtimes` beside the here specified ones will be included in the
                           group. Defaults to None, i.e. not restricted.
            gpu_required: True, if gpu availability is required. Defaults to False.
            min_memory: The minimal amount of memory in MB. Defaults to None, i.e. not restricted.
            min_cpu_cores: The minimum number of cpu cores required. Defaults to None, i.e. not restricted.
            installed_executables: Possibility to only include `Runtimes` that have an specific executables installed.
            filter_commands: Shell commands that can be used for generic filtering.
            working_dir: The directory which shall act as working one. Defaults to None. See the `Runtime` docs for
                         further details.

        Note:
            The filter criteria are passed through the `check_filter()` method of the `Runtime` class. See its
            documentation for further details and examples.

        Returns:
            RuntimeGroup: The created `RuntimeGroup`.
        
        Raises:
            ValueError: Only hosts or excluded_hosts must be provided or Hostname is not contained in the group.
            NoRuntimesError: If no `Runtime` matches the filter criteria or none could be detected.
        """
        self.log.info(f'RuntimeManager starts looking for valid Runtimes.')
        runtimes_dict = self._group.get_runtimes(include_hosts, exclude_hosts)  # Raises ValueError

        final_runtimes = []

        if gpu_required or min_memory or min_cpu_cores or installed_executables or filter_commands:

            self.log.info('RuntimeManager starts evaluating the given filter criteria')

            # Ensure that all runtime info are read asynchronously before printing.
            # Otherwise each runtime will read its information synchronously.
            if gpu_required or min_memory or min_cpu_cores:
                self._group.fill_runtime_info_buffers_async()

            # The evaluation of installed_executables and filter_commands is not yet parallelized.
            # => needs refactoring if heavily used, for now we keep it simple

            for runtime in runtimes_dict.values():
                if runtime.check_filter(gpu_required, min_memory, min_cpu_cores, installed_executables, filter_commands):
                    final_runtimes.append(runtime)

        else:
            final_runtimes = runtimes_dict.values()

        if working_dir:
            for runtime in final_runtimes:
                runtime.working_dir = working_dir

        try:
            group = RuntimeGroup(final_runtimes)
            self.log.debug('RuntimeGroup created by RuntimeManager')
        except ValueError as e:
            raise NoRuntimesDetectedError(e)

        self.log.info(f'RuntimeManager created RuntimeGroup with {str(group.runtime_count)} Runtime(s).')
        return group

    def print_runtime_info(self):
        """Print detailed information of detected `Runtimes` and moreover the names of inactive hosts.

        Note:
            Inactive means that the host is not reachable via ssh or the check vie Runtime.is_valid_runtime() failed.

        """
        print('\n\u001b[1m')
        print(f'{str(self._group.runtime_count)} Runtime(s) and {str(len(self.inactive_hosts))} inactive '
              f'host(s) detected:')
        print('\u001b[0m')

        self._group.print_runtime_info()
        self.print_inactive_hosts()
        print('\n')

    def print_hosts(self):
        """Print detected hosts incl. the inactive ones.

        Note:
            Inactive means that the host is not reachable via ssh or the check vie Runtime.is_valid_runtime() failed.

        """
        print('\n\u001b[1m')
        print(f'{str(self._group.runtime_count)} Runtime(s) and {str(len(self.inactive_hosts))} inactive '
              f'host(s) detected:')
        print('\u001b[0m')

        self._group.print_hosts()
        self.print_inactive_hosts()
        print('\n')

    def print_inactive_hosts(self):
        """Print the inactive hosts.

        Note:
            Inactive means that the host is not reachable via ssh or the check vie Runtime.is_valid_runtime() failed.

        """
        for host in self.inactive_hosts:
            print(f'{host} (inactive)')